# -*- coding: utf-8 -*-
# Part of the PsychoPy library
# Copyright (C) 2012-2020 iSolver Software Solutions (C) 2021 Open Science Tools Ltd.
# Distributed under the terms of the GNU General Public License (GPL).

from psychopy import core
from psychopy.iohub.constants import EventConstants
from psychopy.iohub.client import ioHubConnection

getTime = core.getTime


class Trigger:
    io = None

    def __init__(self, trigger_function=lambda a, b, c: True, user_kwargs={}, repeat_count=0):
        Trigger.io = ioHubConnection.getActiveConnection()
        self.trigger_function = trigger_function
        self.user_kwargs = user_kwargs
        self._last_triggered_event = None
        self._last_triggered_time = None
        self.repeat_count = repeat_count
        self.triggered_count = 0

    def triggered(self, **kwargs):
        if 0 <= self.repeat_count < self.triggered_count:
            return False
        return True

    def getTriggeringEvent(self):
        return self._last_triggered_event

    def getTriggeringTime(self):
        return self._last_triggered_time

    def getTriggeredStateCallback(self):
        return self.trigger_function, self.user_kwargs

    def resetLastTriggeredInfo(self):
        self._last_triggered_event = None
        self._last_triggered_time = None

    def resetTrigger(self):
        self.resetLastTriggeredInfo()
        self.triggered_count = 0

    @classmethod
    def getEventBuffer(cls, copy=False):
        return {}

    @classmethod
    def clearEventHistory(cls, returncopy=False):
        if returncopy:
            return {}

    @classmethod
    def getTriggersFrom(cls, triggers):
        """
        Returns a list of Trigger instances generated based on the contents of the
        input triggers.

        :param triggers:
        :return:
        """
        # Handle different valid trigger object types
        if isinstance(triggers, (list, tuple)):
            # Support is provided for a list of Trigger objects or a list of
            # strings.
            t1 = triggers[0]
            if isinstance(t1, str):
                # triggers is a list of strings, so try and create a list of
                # DeviceEventTrigger's using keyboard device, KEYBOARD_RELEASE
                # event type, and the triggers list elements each as the
                # event.key.
                kbtriggers = []
                for c in triggers:
                    kbtriggers.append(KeyboardTrigger(c, on_press=False))
                trig_list = kbtriggers
            else:
                # Assume triggers is a list of Trigger objects
                trig_list = triggers
        elif isinstance(triggers, (int, float)):
            # triggers is a number, so assume a TimeTrigger is wanted where
            # the delay == triggers. start time will be the fliptime of the
            # last update for drawing to the new target position.
            trig_list = (TimeTrigger(start_time=None, delay=triggers),)
        elif isinstance(triggers, str):
            # triggers is a string, so try and create a
            # DeviceEventTrigger using keyboard device, KEYBOARD_RELEASE
            # event type, and triggers as the event.key.
            trig_list = [KeyboardTrigger(triggers, on_press=False), ]
        elif isinstance(triggers, Trigger):
            # A single Trigger object was provided
            trig_list = (triggers,)
        else:
            raise ValueError('The triggers kwarg could not be understood as a valid triggers input value.')
        return trig_list


class TimeTrigger(Trigger):
    """
    A TimeTrigger associates a delay from the provided start_time
    parameter to when the classes triggered() method returns True.
    start_time and delay can be sec.msec float, or a callable object
    (that takes no parameters).
    """

    def __init__(self, start_time, delay, repeat_count=0, trigger_function=lambda a, b, c: True, user_kwargs={}):
        Trigger.io = ioHubConnection.getActiveConnection()
        Trigger.__init__(self, trigger_function, user_kwargs, repeat_count)

        self._start_time = start_time

        if start_time is None or not callable(start_time):
            def startTimeFunc():
                if self._start_time is None:
                    self._start_time = getTime()
                return self._start_time

            self.startTime = startTimeFunc
        else:
            self.startTime = start_time

        self.delay = delay
        if not callable(delay):
            def delayFunc():
                return delay

            self.delay = delayFunc

    def triggered(self, **kwargs):
        if Trigger.triggered(self) is False:
            return False

        if self.startTime is None:
            start_time = kwargs.get('start_time')
        else:
            start_time = self.startTime()

        if self.delay is None:
            delay = kwargs.get('delay')
        else:
            delay = self.delay()

        ct = getTime()
        if ct - start_time >= delay:
            self._last_triggered_time = ct
            self._last_triggered_event = ct
            self.triggered_count += 1
            return True
        return False

    def resetTrigger(self):
        self.resetLastTriggeredInfo()
        self.triggered_count = 0
        self._start_time = None


class DeviceEventTrigger(Trigger):
    """
    A DeviceEventTrigger associates a set of conditions for a
    DeviceEvent that must be met before the classes triggered() method
    returns True.
    """
    _lastEventsByDevice = dict()

    def __init__(self, device, event_type, event_attribute_conditions={}, repeat_count=-1,
                 trigger_function=lambda a, b, c: True, user_kwargs={}):
        Trigger.io = ioHubConnection.getActiveConnection()
        Trigger.__init__(self, trigger_function, user_kwargs, repeat_count)
        self.device = device
        self.event_type = event_type
        self.event_attribute_conditions = event_attribute_conditions

    def triggered(self, **kwargs):
        if Trigger.triggered(self) is False:
            return False

        events = self.device.getEvents()
        if events is None:
            events = []
        if self.device in self._lastEventsByDevice:
            self._lastEventsByDevice[self.device].extend(events)
        else:
            self._lastEventsByDevice[self.device] = events
        unhandledEvents = self._lastEventsByDevice.get(self.device, [])

        for event in unhandledEvents:
            foundEvent = True
            if event.type != self.event_type:
                foundEvent = False
            else:
                for (attrname, conds) in self.event_attribute_conditions.items():
                    if isinstance(conds, (list, tuple)) and getattr(event, attrname) in conds:
                        # event_value is a list or tuple of possible values
                        # that are OK
                        pass
                    elif getattr(event, attrname) is conds or getattr(event, attrname) == conds:
                        # event_value is a single value
                        pass
                    else:
                        foundEvent = False

            if foundEvent is True:
                self._last_triggered_time = getTime()
                self._last_triggered_event = event
                self.triggered_count += 1
                return True

        return False

    @classmethod
    def getEventBuffer(cls, copy=False):
        if copy:
            return dict(cls._lastEventsByDevice)
        return cls._lastEventsByDevice

    @classmethod
    def clearEventHistory(cls, returncopy=False):
        eventbuffer = None
        if returncopy:
            eventbuffer = dict(cls._lastEventsByDevice)
        cls._lastEventsByDevice.clear()
        return eventbuffer

    def resetLastTriggeredInfo(self):
        Trigger.resetLastTriggeredInfo(self)
        if self.device in self._lastEventsByDevice:
            del self._lastEventsByDevice[self.device]


class KeyboardTrigger(DeviceEventTrigger):
    def __init__(self, key, on_press=False):
        Trigger.io = ioHubConnection.getActiveConnection()
        if on_press:
            etype = EventConstants.KEYBOARD_PRESS
        else:
            etype = EventConstants.KEYBOARD_RELEASE
        DeviceEventTrigger.__init__(self, self.io.devices.keyboard, event_type=etype,
                                    event_attribute_conditions={'key': key})



from psychopy import visual, monitors
import numpy

from psychopy.tests import skip_under_vm, _vmTesting
import pytest


@skip_under_vm(reason="Cannot test gamma in a virtual machine")
def test_low_gamma():
    """setting gamma low (dark screen)"""
    win = visual.Window([600,600], gamma=0.5, autoLog=False)#should make the entire screen bright
    for n in range(5):
        win.flip()
    assert win.useNativeGamma==False
    win.close()


@skip_under_vm(reason="Cannot test gamma in a virtual machine")
def test_mid_gamma():
    """setting gamma high (bright screen)"""
    win = visual.Window([600,600], gamma=2.0, autoLog=False)#should make the entire screen bright
    for n in range(5):
        win.flip()
    assert win.useNativeGamma==False
    win.close()


@skip_under_vm(reason="Cannot test gamma in a virtual machine")
def test_high_gamma():
    """setting gamma high (bright screen)"""
    win = visual.Window([600,600], gamma=4.0, autoLog=False)#should make the entire screen bright
    for n in range(5):
        win.flip()
    assert win.useNativeGamma==False
    win.close()


@skip_under_vm(reason="Cannot test gamma in a virtual machine")
def test_no_gamma():
    """check that no gamma is used if not passed"""
    win = visual.Window([600,600], autoLog=False)#should not change gamma
    assert win.useNativeGamma==True
    win.close()
    """Or if gamma is provided but by a default monitor?"""
    win = visual.Window([600,600], monitor='blaah', autoLog=False)#should not change gamma
    assert win.useNativeGamma==True
    win.close()


@skip_under_vm(reason="Cannot test gamma in a virtual machine")
def test_monitorGetGamma():
    #create our monitor object
    gammaVal = [2.2, 2.2, 2.2]
    mon = monitors.Monitor('test')
    mon.setGamma(gammaVal)
    #create window using that monitor
    win = visual.Window([100,100], monitor=mon, autoLog=False)
    assert numpy.alltrue(win.gamma==gammaVal)
    win.close()


@skip_under_vm(reason="Cannot test gamma in a virtual machine")
def test_monitorGetGammaGrid():
    #create (outdated) gamma grid (new one is [4,6])
    newGrid = numpy.array([[0,150,2.0],#lum
                           [0,30,2.0],#r
                           [0,110,2.0],#g
                           [0,10,2.0]],#b
                           )
    mon = monitors.Monitor('test')
    mon.setGammaGrid(newGrid)
    win = visual.Window([100,100], monitor=mon, autoLog=False)
    assert numpy.alltrue(win.gamma==numpy.array([2.0, 2.0, 2.0]))
    win.close()


@skip_under_vm(reason="Cannot test gamma in a virtual machine")
def test_monitorGetGammaAndGrid():
    """test what happens if gamma (old) and gammaGrid (new) are both present"""
    #create (outdated) gamma grid (new one is [4,6])
    newGrid = numpy.array([[0,150,2.0],#lum
                           [0,30,2.0],#r
                           [0,110,2.0],#g
                           [0,10,2.0]],#b
                           )
    mon = monitors.Monitor('test')
    mon.setGammaGrid(newGrid)
    mon.setGamma([3,3,3])
    #create window using that monitor
    win = visual.Window([100,100], monitor=mon, autoLog=False)
    assert numpy.alltrue(win.gamma==numpy.array([2.0, 2.0, 2.0]))
    win.close()


@skip_under_vm(reason="Cannot test gamma in a virtual machine")
def test_setGammaRamp():
    """test that the gamma ramp is set as requested"""

    testGamma = 2.2

    win = visual.Window([600,600], autoLog=False)
    desiredRamp = numpy.tile(
        visual.gamma.createLinearRamp(
            rampSize=win.backend.getGammaRampSize(),
            driver=win.backend._driver
        ),
        (3, 1)
    )

    if numpy.all(testGamma == 1.0) == False:
        # correctly handles 1 or 3x1 gamma vals
        desiredRamp = desiredRamp**(1.0/numpy.array(testGamma))

    win.gamma = testGamma

    for n in range(5):
        win.flip()

    setRamp = win.backend.getGammaRamp()

    win.close()

    assert numpy.allclose(desiredRamp, setRamp, atol=1.0 / desiredRamp.shape[1])


@skip_under_vm(reason="Cannot test gamma in a virtual machine")
def test_gammaSetGetMatch():
    """test that repeatedly getting and setting the gamma table has no
    cumulative effect."""

    startGammaTable = None

    n_repeats = 2

    for _ in range(n_repeats):

        win = visual.Window([600, 600], autoLog=False)

        for _ in range(5):
            win.flip()

        if startGammaTable is None:
            startGammaTable = win.backend.getGammaRamp()
        else:
            currGammaTable = win.backend.getGammaRamp()

            assert numpy.all(currGammaTable == startGammaTable)

        win.close()


if __name__=='__main__':
    test_high_gamma()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
New backend for the Psychtoolbox portaudio engine
"""
import sys
import os
import time
import re
import weakref

from psychopy import prefs, logging, exceptions
from psychopy.constants import (STARTED, PAUSED, FINISHED, STOPPING,
                                NOT_STARTED)
from .exceptions import SoundFormatError, DependencyError
from ._base import _SoundBase, HammingWindow

try:
    from psychtoolbox import audio
    import psychtoolbox as ptb
except Exception:
    raise DependencyError("psychtoolbox audio failed to import")
try:
    import soundfile as sf
except Exception:
    raise DependencyError("soundfile not working")

import numpy as np

try:
    defaultLatencyClass = int(prefs.hardware['audioLatencyMode'][0])
except (TypeError, IndexError):  # maybe we were given a number instead
    defaultLatencyClass = prefs.hardware['audioLatencyMode']
"""vals in prefs.hardware['audioLatencyMode'] are:
     {0:_translate('Latency not important'),
      1:_translate('Share low-latency driver'),
      2:_translate('Exclusive low-latency'),
      3:_translate('Aggressive low-latency'),
      4:_translate('Latency critical')}
Based on help at http://psychtoolbox.org/docs/PsychPortAudio-Open
"""
# suggestedLatency = 0.005  ## Not currently used. Keep < 1 scr refresh

if prefs.hardware['audioDriver']=='auto':
    audioDriver = None
else:
    audioDriver = prefs.hardware['audioDriver']

if prefs.hardware['audioDevice']=='auto':
    audioDevice = None
else:
    audioDevice = prefs.hardware['audioDevice']

# these will be used by sound.__init__.py
defaultInput = None
defaultOutput = audioDevice


travisCI = bool(str(os.environ.get('TRAVIS')).lower() == 'true')
logging.info("Loaded psychtoolbox audio version {}"
             .format(audio.get_version_info()['version']))

# ask PTB to align verbosity with our current logging level at console
_verbosities = ((logging.DEBUG, 5),
                (logging.INFO, 4),
                (logging.EXP, 3),
                (logging.WARNING, 2),
                (logging.ERROR, 1))

for _logLevel, _verbos in _verbosities:
    if logging.console.level <= _logLevel:
        audio.verbosity(_verbos)
        break


def init(rate=48000, stereo=True, buffer=128):
    pass  # for compatibility with other backends


def getDevices(kind=None):
    """Returns a dict of dict of audio devices of specified `kind`

    kind can be None, 'input' or 'output'
    The dict keys are names, and items are dicts of properties
    """
    if sys.platform=='win32':
        deviceTypes = 13  # only WASAPI drivers need apply!
    else:
        deviceTypes = None
    devs = {}
    if travisCI:  # travis-CI testing does not have a sound device
        return devs
    else:
        allDevs = audio.get_devices(device_type=deviceTypes)

    # annoyingly query_devices is a DeviceList or a dict depending on number
    if type(allDevs) == dict:
        allDevs = [allDevs]

    for ii, dev in enumerate(allDevs):
        if kind and kind.startswith('in'):
            if dev['NrInputChannels'] < 1:
                continue
        elif kind and kind.startswith('out'):
            if dev['NrOutputChannels'] < 1:
                continue
        # we have a valid device so get its name
        # newline characters must be removed
        devName = dev['DeviceName'].replace('\r\n', '')
        devs[devName] = dev
        dev['id'] = ii
    return devs


def getStreamLabel(sampleRate, channels, blockSize):
    """Returns the string repr of the stream label
    """
    return "{}_{}_{}".format(sampleRate, channels, blockSize)


class _StreamsDict(dict):
    """Keeps track of what streams have been created. On macOS we can have
    multiple streams under portaudio but under windows we can only have one.

    use the instance `streams` rather than creating a new instance of this
    """

    def getStream(self, sampleRate, channels, blockSize):
        """Gets a stream of exact match or returns a new one
        (if possible for the current operating system)
        """
        # if the query looks flexible then try getSimilar
        if channels == -1 or blockSize == -1:
            return self._getSimilar(sampleRate,
                                    channels=channels,
                                    blockSize=blockSize)
        else:
            return self._getStream(sampleRate,
                                   channels=channels,
                                   blockSize=blockSize)

    def _getSimilar(self, sampleRate, channels=-1, blockSize=-1):
        """Do we already have a compatible stream?

        Many sounds can allow channels and blocksize to change but samplerate
        is generally fixed. Any values set to -1 above will be flexible. Any
        values set to an alternative number will be fixed

        usage:

            label, stream = streams._getSimilar(sampleRate=44100,  # must match
                                               channels=-1,  # any
                                               blockSize=-1)  # wildcard
        """
        label = getStreamLabel(sampleRate, channels, blockSize)
        # replace -1 with any regex integer
        simil = re.compile(label.replace("-1", r"[-+]?(\d+)"))  # I hate REGEX!
        for thisFormat in self:
            if simil.match(thisFormat):  # we found a close-enough match
                return thisFormat, self[thisFormat]
        # if we've been given values in each place then create stream
        if (sampleRate not in [None, -1, 0] and
                channels not in [None, -1] and
                blockSize not in [None, -1]):
            return self._getStream(sampleRate, channels, blockSize)

    def _getStream(self, sampleRate, channels, blockSize):
        """Strict check for this format or create new
        """
        label = getStreamLabel(sampleRate, channels, blockSize)
        # try to retrieve existing stream of that name
        if label in self:
            pass
        # todo: check if this is still needed on win32
        # on some systems more than one stream isn't supported so check
        elif sys.platform == 'win32' and len(self):
            raise SoundFormatError(
                "Tried to create audio stream {} but {} already exists "
                "and {} doesn't support multiple portaudio streams"
                    .format(label, list(self.keys())[0], sys.platform)
            )
        else:

            # create new stream
            self[label] = _MasterStream(sampleRate, channels, blockSize,
                                       device=defaultOutput)
        return label, self[label]


streams = _StreamsDict()


class _MasterStream(audio.Stream):
    def __init__(self, sampleRate, channels, blockSize,
                 device=None, duplex=False, mode=1,
                 audioLatencyClass=None):
        # initialise thread
        if audioLatencyClass is None:
            audioLatencyClass = defaultLatencyClass
        self.streamLabel = None
        self.streams = []
        self.list = []
        # sound stream info
        self.sampleRate = sampleRate
        self.channels = channels
        self.duplex = duplex
        self.blockSize = blockSize
        self.label = getStreamLabel(sampleRate, channels, blockSize)
        if type(device) == list and len(device):
            device = device[0]
        if type(device)==str:  # we need to convert name to an ID or make None
            devs = getDevices('output')
            if device in devs:
                deviceID = devs[device]['DeviceIndex']
            else:
                deviceID = None
        else:
            deviceID = device
        self.sounds = []  # list of dicts for sounds currently playing
        self.takeTimeStamp = False
        self.frameN = 1
        # self.frameTimes = range(5)  # DEBUGGING: store the last 5 callbacks
        if not travisCI:  # travis-CI testing does not have a sound device
            try:
                audio.Stream.__init__(self, device_id=deviceID, mode=mode+8,
                                    latency_class=audioLatencyClass,
                                    freq=sampleRate, 
                                    channels=channels,
                                    )  # suggested_latency=suggestedLatency
            except OSError as e:
                audio.Stream.__init__(self, device_id=deviceID, mode=mode+8,
                                    latency_class=audioLatencyClass,
                                    # freq=sampleRate, 
                                    channels=channels,
                                    )
                self.sampleRate = self.status['SampleRate']
                print("Failed to start PTB.audio with requested rate of "
                      "{} but succeeded with a default rate ({}). "
                      "This is depends on the selected latency class and device."
                      .format(sampleRate, self.sampleRate))
            except TypeError as e:
                print("device={}, mode={}, latency_class={}, freq={}, channels={}"
                      .format(device, mode+8, audioLatencyClass, sampleRate, channels))
                raise(e)
            except Exception as e:
                audio.Stream.__init__(self, mode=mode+8,
                                    latency_class=audioLatencyClass,
                                    freq=sampleRate, 
                                    channels=channels,
                                    )
                
                if "there isn't any audio output device" in str(e):
                    print("Failed to load audio device:\n"
                          "    '{}'\n"
                          "so fetching default audio device instead: \n"
                          "    '{}'"
                          .format(device, 'test'))
            self.start(0, 0, 1)
            # self.device = self._sdStream.device
            # self.latency = self._sdStream.latency
            # self.cpu_load = self._sdStream.cpu_load
        self._tSoundRequestPlay = 0


class SoundPTB(_SoundBase):
    """Play a variety of sounds using the new PsychPortAudio library
    """

    def __init__(self, value="C", secs=0.5, octave=4, stereo=-1,
                 volume=1.0, loops=0,
                 sampleRate=None, blockSize=128,
                 preBuffer=-1,
                 hamming=True,
                 startTime=0, stopTime=-1,
                 name='', autoLog=True,
                 syncToWin=None):
        """
        :param value: note name ("C","Bfl"), filename or frequency (Hz)
        :param secs: duration (for synthesised tones)
        :param octave: which octave to use for note names (4 is middle)
        :param stereo: -1 (auto), True or False
                        to force sounds to stereo or mono
        :param volume: float 0-1
        :param loops: number of loops to play (-1=forever, 0=single repeat)
        :param sampleRate: sample rate for synthesized tones
        :param blockSize: the size of the buffer on the sound card
                         (small for low latency, large for stability)
        :param preBuffer: integer to control streaming/buffering
                           - -1 means store all
                           - 0 (no buffer) means stream from disk
                           - potentially we could buffer a few secs(!?)
        :param hamming: boolean (default True) to indicate if the sound should
                        be apodized (i.e., the onset and offset smoothly ramped up from
                        down to zero). The function apodize uses a Hanning window, but
                        arguments named 'hamming' are preserved so that existing code
                        is not broken by the change from Hamming to Hanning internally.
                        Not applied to sounds from files.
        :param startTime: for sound files this controls the start of snippet
        :param stopTime: for sound files this controls the end of snippet
        :param name: string for logging purposes
        :param autoLog: whether to automatically log every change
        :param syncToWin: if you want start/stop to sync with win flips add this
        """
        self.sound = value
        self.name = name
        self.secs = secs  # for any synthesised sounds (notesand freqs)
        self.octave = octave  # for note name sounds
        self.loops = self._loopsRequested = loops
        self._loopsFinished = 0
        self.volume = volume
        self.startTime = startTime  # for files
        self.stopTime = stopTime  # for files specify thesection to be played
        self.blockSize = blockSize  # can be per-sound unlike other backends
        self.preBuffer = preBuffer
        self.frameN = 0
        self._tSoundRequestPlay = 0
        self.sampleRate = sampleRate
        self.channels = None  # let this be set by stereo
        self.stereo = stereo
        self.duplex = None
        self.autoLog = autoLog
        self.streamLabel = ""
        self.sourceType = 'unknown'  # set to be file, array or freq
        self.sndFile = None
        self.sndArr = None
        self.hamming = hamming
        self._hammingWindow = None  # will be created during setSound
        self.win=syncToWin
        # setSound (determines sound type)
        self.setSound(value, secs=self.secs, octave=self.octave,
                      hamming=self.hamming)
        self.status = NOT_STARTED

    def _getDefaultSampleRate(self):
        """Check what streams are open and use one of these"""
        if len(streams):
            return list(streams.values())[0].sampleRate
        else:
            return 48000  # seems most widely supported

    @property
    def statusDetailed(self):
        if not self.track:
            return None
        return self.track.status

    @property
    def status(self):
        """status gives a simple value from psychopy.constants to indicate
        NOT_STARTED, STARTED, FINISHED, PAUSED

        Psychtoolbox sounds also have a statusDetailed property with further info"""
        if self.__dict__['status']==STARTED:
            # check portaudio to see if still playing
            pa_status = self.statusDetailed
            if not pa_status['Active'] and pa_status['State']==0:
                # we were playing and now not so presumably FINISHED
                self._EOS()

        return self.__dict__['status']


    @status.setter
    def status(self, newStatus):
        self.__dict__['status'] = newStatus

    @property
    def volume(self):
        return self.__dict__['volume']

    @volume.setter
    def volume(self, newVolume):
        self.__dict__['volume'] = newVolume
        if 'track' in self.__dict__:
            # Update volume of an existing track, if it exists.
            # (BUGFIX, otherwise only the member variable is updated, but the sound
            # volume does not change while playing - Suddha Sourav, 14.10.2020)
            self.__dict__['track']().volume = newVolume
        else:
            return None

    @property
    def stereo(self):
        return self.__dict__['stereo']

    @stereo.setter
    def stereo(self, val):
        self.__dict__['stereo'] = val
        if val == True:
            self.__dict__['channels'] = 2
        elif val == False:
            self.__dict__['channels'] = 1
        elif val == -1:
            self.__dict__['channels'] = -1

    def setSound(self, value, secs=0.5, octave=4, hamming=None, log=True):
        """Set the sound to be played.

        Often this is not needed by the user - it is called implicitly during
        initialisation.

        :parameters:

            value: can be a number, string or an array:
                * If it's a number between 37 and 32767 then a tone will
                  be generated at that frequency in Hz.
                * It could be a string for a note ('A', 'Bfl', 'B', 'C',
                  'Csh'. ...). Then you may want to specify which octave.
                * Or a string could represent a filename in the current
                  location, or mediaLocation, or a full path combo
                * Or by giving an Nx2 numpy array of floats (-1:1) you can
                  specify the sound yourself as a waveform

            secs: duration (only relevant if the value is a note name or
                a frequency value)

            octave: is only relevant if the value is a note name.
                Middle octave of a piano is 4. Most computers won't
                output sounds in the bottom octave (1) and the top
                octave (8) is generally painful
        """
        # reset self.loops to what was requested (in case altered for infinite play of tones)
        self.loops = self._loopsRequested
        # start with the base class method
        _SoundBase.setSound(self, value, secs, octave, hamming, log)

    def _setSndFromFile(self, filename):
        self.sndFile = f = sf.SoundFile(filename)
        self.sourceType = 'file'
        self.sampleRate = f.samplerate
        if self.channels == -1:  # if channels was auto then set to file val
            self.channels = f.channels
        fileDuration = float(len(f)) / f.samplerate  # needed for duration?
        # process start time
        if self.startTime and self.startTime > 0:
            startFrame = self.startTime * self.sampleRate
            self.sndFile.seek(int(startFrame))
            self.t = self.startTime
        else:
            self.t = 0
        # process stop time
        if self.stopTime and self.stopTime > 0:
            requestedDur = self.stopTime - self.t
            self.duration = min(requestedDur, fileDuration)
        else:
            self.duration = fileDuration - self.t
        # can now calculate duration in frames
        self.durationFrames = int(round(self.duration * self.sampleRate))
        # are we preloading or streaming?
        if self.preBuffer == 0:
            # no buffer - stream from disk on each call to nextBlock
            pass
        elif self.preBuffer == -1:
            # full pre-buffer. Load requested duration to memory
            sndArr = self.sndFile.read(
                frames=int(self.sampleRate * self.duration))
            self.sndFile.close()
            self._setSndFromArray(sndArr)
        self._channelCheck(
            self.sndArr)  # Check for fewer channels in stream vs data array


    def _setSndFromArray(self, thisArray):

        self.sndArr = np.asarray(thisArray).astype('float32')
        if thisArray.ndim == 1:
            self.sndArr.shape = [len(thisArray), 1]  # make 2D for broadcasting
        if self.channels == 2 and self.sndArr.shape[1] == 1:  # mono -> stereo
            self.sndArr = self.sndArr.repeat(2, axis=1)
        elif self.sndArr.shape[1] == 1:  # if channels in [-1,1] then pass
            pass
        else:
            try:
                self.sndArr.shape = [len(thisArray), 2]
            except ValueError:
                raise ValueError("Failed to format sound with shape {} "
                                 "into sound with channels={}"
                                 .format(self.sndArr.shape, self.channels))

        # is this stereo?
        if self.stereo == -1:  # auto stereo. Try to detect
            if self.sndArr.shape[1] == 1:
                self.stereo = 0
            elif self.sndArr.shape[1] == 2:
                self.stereo = 1
            else:
                raise IOError("Couldn't determine whether array is "
                              "stereo. Shape={}".format(self.sndArr.shape))
        self._nSamples = thisArray.shape[0]
        if self.stopTime == -1:
            self.duration = self._nSamples / float(self.sampleRate)
        # set to run from the start:
        self.seek(0)
        self.sourceType = "array"

        if not self.track:  # do we have one already?
            self.track = audio.Slave(self.stream.handle, data=self.sndArr,
                                     volume=self.volume)
        else:
            self.track.stop()
            self.track.fill_buffer(self.sndArr)

    def _channelCheck(self, array):
        """Checks whether stream has fewer channels than data. If True, ValueError"""
        if self.channels < array.shape[1]:
            msg = (
                "The sound stream is set up incorrectly. You have fewer channels in the buffer "
                "than in data file ({} vs {}).\n**Ensure you have selected 'Force stereo' in "
                "experiment settings**".format(self.channels, array.shape[1]))
            logging.error(msg)
            raise ValueError(msg)

    def play(self, loops=None, when=None, log=True):
        """Start the sound playing
        """
        if loops is not None and self.loops != loops:
            self.setLoops(loops)
        self.status = STARTED
        self._tSoundRequestPlay = time.time()

        if hasattr(when, 'getFutureFlipTime'):
            logTime = when.getFutureFlipTime(clock=None)
            when = when.getFutureFlipTime(clock='ptb')
        elif when is None and hasattr(self.win, 'getFutureFlipTime'):
            logTime = self.win.getFutureFlipTime(clock=None)
            when = self.win.getFutureFlipTime(clock='ptb')
        else:
            logTime = None
        self.track.start(repetitions=loops, when=when)
        # time.sleep(0.)
        if log and self.autoLog:
            logging.exp(u"Sound %s started" % (self.name), obj=self, t=logTime)

    def pause(self):
        """Stop the sound but play will continue from here if needed
        """
        self.status = PAUSED
        self.track.stop()

    def stop(self, reset=True, log=True):
        """Stop the sound and return to beginning
        """
        if self.status == FINISHED:
            return
        self.track.stop()
        if reset:
            self.seek(0)
        if log and self.autoLog:
            logging.exp(u"Sound %s stopped" % (self.name), obj=self)
        self.status = FINISHED

    def seek(self, t):
        self.t = t
        self.frameN = int(round(t * self.sampleRate))
        if self.sndFile and not self.sndFile.closed:
            self.sndFile.seek(self.frameN)

    def _EOS(self, reset=True, log=True):
        """Function called on End Of Stream
        """
        self.status = STOPPING
        self._loopsFinished += 1
        if self.loops == 0:
            self.stop(reset=reset, log=False)
        elif self.loops > 0 and self._loopsFinished >= self.loops:
            self.stop(reset=reset, log=False)
        if log and self.autoLog:
            logging.exp(u"Sound %s reached end of file" % (self.name), obj=self)

    @property
    def stream(self):
        """Read-only property returns the the stream on which the sound
        will be played
        """
        if not self.streamLabel:
            try:
                label, s = streams.getStream(sampleRate=self.sampleRate,
                                             channels=self.channels,
                                             blockSize=self.blockSize)
            except SoundFormatError as err:
                # try to use something similar (e.g. mono->stereo)
                # then check we have an appropriate stream open
                altern = streams._getSimilar(sampleRate=self.sampleRate,
                                             channels=-1,
                                             blockSize=-1)
                if altern is None:
                    raise SoundFormatError(err)
                else:  # safe to extract data
                    label, s = altern
                # update self in case it changed to fit the stream
                self.sampleRate = s.sampleRate
                self.channels = s.channels
                self.blockSize = s.blockSize
            self.streamLabel = label

        return streams[self.streamLabel]

    def __del__(self):
        if self.track:
            self.track.close()
        self.track = None

    @property
    def track(self):
        """The track on the master stream to which we belong"""
        # the track is actually a weak reference to avoid circularity
        if 'track' in self.__dict__:
            return self.__dict__['track']()
        else:
            return None

    @track.setter
    def track(self, track):
        if track is None:
            self.__dict__['track'] = None
        else:
            self.__dict__['track'] = weakref.ref(track)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Demo of MovieStim

Different systems have different sets of codecs.
avbin (which PsychoPy uses to load movies) seems not to load compressed audio on all systems.
To create a movie that will play on all systems I would recommend using the format:
    video: H.264 compressed,
    audio: Linear PCM
"""
from psychopy import visual, core, event, constants

win = visual.Window((800, 600))
mov = visual.MovieStim3(win, 'jwpIntro.mp4', size=(320, 240),
    flipVert=False, flipHoriz=False, loop=False)
print('orig movie size=%s' % mov.size)
print('duration=%.2fs' % mov.duration)
globalClock = core.Clock()

while mov.status != constants.FINISHED:
    mov.draw()
    win.flip()
    if event.getKeys():
        break

win.close()
core.quit()

# The contents of this file are in the public domain.



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Example for reading events from an iohub hdf5 file. Events can optionally be grouped (typically into trials) by either:
    1. Reading iohub experiment messages
    2. Reading the iohub .hdf5 condition variables table
    3. Reading a psychopy trial-by-trial .csv data (results) file
When grouping events, use the TRIAL_START and TRIAL_END variables to specify the message text or column names to use
to find the start and end time for each trial period.

SAVE_EVENT_TYPE and SAVE_EVENT_FIELDS specify the event type, and which event fields, are saved. Set to None
to be prompted for the event type.

Each event is saved as a row in a tab delimited text file.
"""
from psychopy.iohub.datastore.util import saveEventReport

# Specify the iohub .hdf5 file to process. None will prompt for file selection when script is run.
IOHUB_DATA_FILE = None

# If True, psychopy .csv file with same path as IOHUB_DATA_FILE will be used
USE_PSYCHOPY_DATA_FILE = False

# If True, iohub .hdf5 condition variable table will be used to split events based on TRIAL_START and TRIAL_END
USE_CONDITIONS_TABLE = False

# Specify the experiment message text used to split events into trial periods.
# Set both to None to save all events.
TRIAL_START = None#'text.started' #  'target.started'
TRIAL_END = None#'fix_end_stim.started' #  'fix_end_stim.started'

# Specify which event type to save. Setting to None will prompt to select an event table
SAVE_EVENT_TYPE = None  # 'MonocularEyeSampleEvent'
# Specify which event fields to save. Setting to None will save all event fields.
SAVE_EVENT_FIELDS = None  # ['time', 'gaze_x', 'gaze_y', 'pupil_measure1', 'status']

# SAVE_EVENT_TYPE = 'BinocularEyeSampleEvent'
# SAVE_EVENT_FIELDS = ['time', 'left_gaze_x', 'left_gaze_y', 'left_pupil_measure1',
#                     'right_gaze_x', 'right_gaze_y', 'right_pupil_measure1', 'status']

# SAVE_EVENT_TYPE = 'MessageEvent'
# SAVE_EVENT_FIELDS = ['time', 'text']

if __name__ == '__main__':
    result = saveEventReport(hdf5FilePath=IOHUB_DATA_FILE,
                             eventType=SAVE_EVENT_TYPE,
                             eventFields=SAVE_EVENT_FIELDS,
                             useConditionsTable=USE_CONDITIONS_TABLE,
                             usePsychopyDataFile=USE_PSYCHOPY_DATA_FILE,
                             trialStart=TRIAL_START,
                             trialStop=TRIAL_END)
    if result:
        file_saved, events_saved = result
        print("Saved %d events to %s." % (events_saved, file_saved))
    else:
        raise RuntimeError("saveEventReport failed.")



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Created on Tue Apr 23 11:28:32 2013

Provides the high resolution timebase used by psychopy, and defines some time
related utility Classes.

Moved functionality from core.py so a common code
base could be used in core.py and logging.py; vs. duplicating the getTime and
Clock logic.

@author: Sol
@author: Jon
"""

# Part of the PsychoPy library
# Copyright (C) 2002-2018 Jonathan Peirce (C) 2019-2022 Open Science Tools Ltd.
# Distributed under the terms of the GNU General Public License (GPL).

import time
import sys
from pkg_resources import parse_version

try:
    import pyglet
except ImportError:
    pass  # pyglet is not installed

from psychopy.constants import STARTED, NOT_STARTED, FINISHED
import psychopy.logging  # Absolute import to work around circularity

# set the default timing mechanism
getTime = None

# Select the timer to use as the psychopy high resolution time base. Selection
# is based on OS and Python version.
#
# Three requirements exist for the psychopy time base implementation:
#     A) The Python interpreter does not apply an offset to the times returned
#        based on when the timer module being used was loaded or when the
#        timer function first called was first called.
#     B) The timer implementation used must be monotonic and report elapsed
#        time between calls, 'not' system or CPU usage time.
#     C) The timer implementation must provide a resolution of 50 usec or
#        better.
#
# Given the above requirements, psychopy selects a timer implementation as
# follows:
#     1) On Windows, the Windows Query Performance Counter API is used using
#        ctypes access.
#     2) On other OS's, if the Python version being used is 2.6 or lower,
#        time.time is used. For Python 2.7 and above, the timeit.default_timer
#        function is used.
try:
    import psychtoolbox
    havePTB = True
except ImportError:
    havePTB = False

if havePTB:
    # def getTime():
    #     secs, wallTime, error = psychtoolbox.GetSecs('allclocks')
    #     return wallTime
    getTime = psychtoolbox.GetSecs
elif sys.platform == 'win32':
    from ctypes import byref, c_int64, windll
    _fcounter = c_int64()
    _qpfreq = c_int64()
    windll.Kernel32.QueryPerformanceFrequency(byref(_qpfreq))
    _qpfreq = float(_qpfreq.value)
    _winQPC = windll.Kernel32.QueryPerformanceCounter

    def getTime():
        _winQPC(byref(_fcounter))
        return _fcounter.value / _qpfreq

elif sys.platform == "darwin":
    # Monotonic getTime with absolute origin. Suggested by @aforren1, and
    # copied from github.com/aforren1/toon/blob/master/toon/input/mac_clock.py 
    import ctypes
    _libc = ctypes.CDLL('/usr/lib/libc.dylib', use_errno=True)

    # create helper class to store data
    class mach_timebase_info_data_t(ctypes.Structure):
        _fields_ = (('numer', ctypes.c_uint32),
                    ('denom', ctypes.c_uint32))

    # get function and set response type
    _mach_absolute_time = _libc.mach_absolute_time
    _mach_absolute_time.restype = ctypes.c_uint64

    # calculate timebase
    _timebase = mach_timebase_info_data_t()
    _libc.mach_timebase_info(ctypes.byref(_timebase))
    _ticks_per_second = _timebase.numer / _timebase.denom * 1.0e9

    # scaling factor so that timing works correctly on Intal and Apple Silicon
    _scaling_factor = _timebase.numer / _timebase.denom

    # then define getTime func
    def getTime():
        return (_mach_absolute_time() * _scaling_factor) / 1.0e9

else:
    import timeit
    getTime = timeit.default_timer


class MonotonicClock:
    """A convenient class to keep track of time in your experiments using a
    sub-millisecond timer.

    Unlike the :class:`~psychopy.core.Clock` this cannot be reset to
    arbitrary times. For this clock t=0 always represents the time that
    the clock was created.

    Don't confuse this `class` with `core.monotonicClock` which is an
    `instance` of it that got created when PsychoPy.core was imported.
    That clock instance is deliberately designed always to return the
    time since the start of the study.

    Version Notes: This class was added in PsychoPy 1.77.00

    """
    def __init__(self, start_time=None):
        super(MonotonicClock, self).__init__()
        if start_time is None:
            # this is sub-millisecond timer in python
            self._timeAtLastReset = getTime()
        else:
            self._timeAtLastReset = start_time

    def getTime(self, applyZero=True):
        """Returns the current time on this clock in secs (sub-ms precision).

        If applying zero then this will be the time since the clock was created
        (typically the beginning of the script).

        If not applying zero then it is whatever the underlying clock uses as
        its base time but that is system dependent. e.g. can be time since
        reboot, time since Unix Epoch etc
        """
        if applyZero:
            return getTime() - self._timeAtLastReset
        else:
            return getTime()

    def getLastResetTime(self):
        """
        Returns the current offset being applied to the high resolution
        timebase used by Clock.
        """
        return self._timeAtLastReset


monotonicClock = MonotonicClock()


class Clock(MonotonicClock):
    """A convenient class to keep track of time in your experiments.
    You can have as many independent clocks as you like (e.g. one
    to time responses, one to keep track of stimuli ...)

    This clock is identical to the :class:`~psychopy.core.MonotonicClock`
    except that it can also be reset to 0 or another value at any point.

    """
    def __init__(self):
        super(Clock, self).__init__()

    def reset(self, newT=0.0):
        """Reset the time on the clock. With no args time will be
        set to zero. If a float is received this will be the new
        time on the clock
        """
        self._timeAtLastReset = getTime() + newT

    def add(self, t):
        """Add more time to the clock's 'start' time (t0).

        Note that, by adding time to t0, you make the current time
        appear less. Can have the effect that getTime() returns a negative
        number that will gradually count back up to zero.

        e.g.::

            timer = core.Clock()
            timer.add(5)
            while timer.getTime()<0:
                # do something
        """
        self._timeAtLastReset += t


class CountdownTimer(Clock):
    """Similar to a :class:`~psychopy.core.Clock` except that time counts down
    from the time of last reset.

    Parameters
    ----------
    start : float or int
        Starting time in seconds to countdown on.

    Examples
    --------
    Create a countdown clock with a 5 second duration::

        timer = core.CountdownTimer(5)
        while timer.getTime() > 0:  # after 5s will become negative
            # do stuff

    """
    def __init__(self, start=0):
        super(CountdownTimer, self).__init__()
        self._countdown_duration = start
        if start:
            self.add(start)

    def getTime(self):
        """Returns the current time left on this timer in seconds with sub-ms
        precision (`float`).
        """
        return self._timeAtLastReset - getTime()

    def reset(self, t=None):
        """Reset the time on the clock.

        Parameters
        ----------
        t : float, int or None
            With no args (`None`), time will be set to the time used for last
            reset (or start time if no previous resets). If a number is
            received, this will be the new time on the clock.

        """
        if t is None:
            Clock.reset(self, self._countdown_duration)
        else:
            self._countdown_duration = t
            Clock.reset(self, t)


class StaticPeriod:
    """A class to help insert a timing period that includes code to be run.

    Parameters
    ----------
    screenHz : int or None
    the frame rate of the monitor (leave as None if you
            don't want this accounted for)
    win : :class:`~psychopy.visual.Window`
        If a :class:`~psychopy.visual.Window` is given then
        :class:`StaticPeriod` will also pause/restart frame interval recording.
    name : str
        Give this StaticPeriod a name for more informative logging messages.

    Examples
    --------
    Typical usage for the static period::

        fixation.draw()
        win.flip()
        ISI = StaticPeriod(screenHz=60)
        ISI.start(0.5)  # start a period of 0.5s
        stim.image = 'largeFile.bmp'  # could take some time
        ISI.complete()  # finish the 0.5s, taking into account one 60Hz frame

        stim.draw()
        win.flip()  # the period takes into account the next frame flip
        # time should now be at exactly 0.5s later than when ISI.start()
        # was called

    """
    def __init__(self, screenHz=None, win=None, name='StaticPeriod'):
        self.status = NOT_STARTED
        self.countdown = CountdownTimer()
        self.name = name
        self.win = win

        if screenHz is None:
            self.frameTime = 0
        else:
            self.frameTime = 1.0 / screenHz

        self._winWasRecordingIntervals = False

    def start(self, duration):
        """Start the period. If this is called a second time, the timer will
        be reset and starts again

        Parameters
        ----------
        duration : float or int
            The duration of the period, in seconds.

        """
        self.status = STARTED
        self.countdown.reset(duration - self.frameTime)

        # turn off recording of frame intervals throughout static period
        if self.win:
            self._winWasRecordingIntervals = self.win.recordFrameIntervals
            self.win.recordFrameIntervals = False

    def complete(self):
        """Completes the period, using up whatever time is remaining with a
        call to `wait()`.

        Returns
        -------
        float
            `1` for success, `0` for fail (the period overran).

        """
        self.status = FINISHED
        timeRemaining = self.countdown.getTime()

        if self.win:
            self.win.recordFrameIntervals = self._winWasRecordingIntervals

        if timeRemaining < 0:
            msg = ('We overshot the intended duration of %s by %.4fs. The '
                   'intervening code took too long to execute.')
            vals = self.name, abs(timeRemaining)
            psychopy.logging.warn(msg % vals)

            return 0

        wait(timeRemaining)

        return 1


def wait(secs, hogCPUperiod=0.2):
    """Wait for a given time period.

    If `secs=10` and `hogCPU=0.2` then for 9.8s Python's `time.sleep` function
    will be used, which is not especially precise, but allows the cpu to
    perform housekeeping. In the final `hogCPUperiod` the more precise
    method of constantly polling the clock is used for greater precision.

    If you want to obtain key-presses during the wait, be sure to use
    pyglet and to hogCPU for the entire time, and then call
    :func:`psychopy.event.getKeys()` after calling
    :func:`~.psychopy.core.wait()`

    If you want to suppress checking for pyglet events during the wait, do this
    once::

        core.checkPygletDuringWait = False

    and from then on you can do::

        core.wait(sec)

    This will preserve terminal-window focus during command line usage.

    Parameters
    ----------
    secs : float or int
    hogCPUperiod : float or int

    """
    from . import core

    # initial relaxed period, using sleep (better for system resources etc)
    if secs > hogCPUperiod:
        time.sleep(secs - hogCPUperiod)
        secs = hogCPUperiod  # only this much is now left

    # hog the cpu, checking time
    t0 = getTime()
    while (getTime() - t0) < secs:
        if not (core.havePyglet and core.checkPygletDuringWait):
            continue
        # let's see if pyglet collected any event in meantime
        try:
            # this takes focus away from command line terminal window:
            if parse_version(pyglet.version) < parse_version('1.2'):
                # events for sounds/video should run independently of wait()
                pyglet.media.dispatch_events()
        except AttributeError:
            # see http://www.pyglet.org/doc/api/pyglet.media-module.html#dispatch_events
            # Deprecated: Since pyglet 1.1, Player objects schedule themselves
            # on the default clock automatically. Applications should not call
            # pyglet.media.dispatch_events().
            pass
        for winWeakRef in core.openWindows:
            win = winWeakRef()
            if (win.winType == "pyglet" and
                    hasattr(win.winHandle, "dispatch_events")):
                win.winHandle.dispatch_events()  # pump events


def getAbsTime():
    """Get the absolute time.

    This uses the same clock-base as the other timing features, like
    `getTime()`. The time (in seconds) ignores the time-zone (like `time.time()`
    on linux). To take the timezone into account, use
    `int(time.mktime(time.gmtime()))`.

    Absolute times in seconds are especially useful to add to generated file
    names for being unique, informative (= a meaningful time stamp), and because
    the resulting files will always sort as expected when sorted in
    chronological, alphabetical, or numerical order, regardless of locale and so
    on.

    Version Notes: This method was added in PsychoPy 1.77.00

    Returns
    -------
    float
        Absolute Unix time (i.e., whole seconds elapsed since Jan 1, 1970).

    """
    return int(time.mktime(time.localtime()))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Part of the PsychoPy library
# Copyright (C) 2002-2018 Jonathan Peirce (C) 2019-2022 Open Science Tools Ltd.
# Distributed under the terms of the GNU General Public License (GPL).

"""Converting code parameters and components from python (PsychoPy)
to JS (ES6/PsychoJS)
"""

import ast
from pathlib import Path

import astunparse
import esprima
from os import path
from psychopy import logging

from io import StringIO
from psychopy.experiment.py2js_transpiler import translatePythonToJavaScript


class NamesJS(dict):
    def __getitem__(self, name):
        try:
            return dict.__getitem__(self, name)
        except:
            return "{}".format(name)


namesJS = NamesJS()
namesJS['sin'] = 'Math.sin'
namesJS['cos'] = 'Math.cos'
namesJS['tan'] = 'Math.tan'
namesJS['pi'] = 'Math.PI'
namesJS['rand'] = 'Math.random'
namesJS['random'] = 'Math.random'
namesJS['sqrt'] = 'Math.sqrt'
namesJS['abs'] = 'Math.abs'
namesJS['randint'] = 'util.randint'
namesJS['round'] = 'util.round'  # better than Math.round, supports n DPs arg
namesJS['sum'] = 'util.sum'


class TupleTransformer(ast.NodeTransformer):
    """ An ast subclass that walks the abstract syntax tree and
    allows modification of nodes.

    This class transforms a tuple to a list.

    :returns node
    """
    def visit_Tuple(self, node):
        return ast.List(node.elts, node.ctx)

class Unparser(astunparse.Unparser):
    """astunparser had buried the future_imports option underneath its init()
    so we need to override that method and change it."""

    def __init__(self, tree, file):
        """Unparser(tree, file=sys.stdout) -> None.
         Print the source for tree to file."""
        self.f = file
        self.future_imports = ['unicode_literals']
        self._indent = 0
        self.dispatch(tree)
        self.f.flush()


def unparse(tree):
    v = StringIO()
    Unparser(tree, file=v)
    return v.getvalue()


def expression2js(expr):
    """Convert a short expression (e.g. a Component Parameter) Python to JS"""

    # if the code contains a tuple (anywhere), convert parenths to be list.
    # This now works for compounds like `(2*(4, 5))` where the inner
    # parenths becomes a list and the outer parens indicate priority.
    # This works by running an ast transformer class to swap the contents of the tuple
    # into a list for the number of tuples in the expression.
    try:
        syntaxTree = ast.parse(expr)
    except Exception:
        try:
            syntaxTree = ast.parse(str(expr))
        except Exception as err:
            logging.error(err)
            return

    for node in ast.walk(syntaxTree):
        TupleTransformer().visit(node)  # Transform tuples to list
        # for py2 using 'unicode_literals' we don't want
        if isinstance(node, ast.Str) and type(node.s)==bytes:
            node.s = str(node.s, 'utf-8')
        elif isinstance(node, ast.Str) and node.s.startswith("u'"):
            node.s = node.s[1:]
        if isinstance(node, ast.Name):
            if node.id == 'undefined':
                continue
            node.id = namesJS[node.id]
    jsStr = unparse(syntaxTree).strip()
    if not any(ch in jsStr for ch in ("=",";","\n")):
        try:
            jsStr = translatePythonToJavaScript(jsStr)
            if jsStr.endswith(';\n'):
                jsStr = jsStr[:-2]
        except:
            # If translation fails, just use old translation
            pass
    return jsStr


def snippet2js(expr):
    """Convert several lines (e.g. a Code Component) Python to JS"""
    # for now this is just adding ';' onto each line ending so will fail on
    # most code (e.g. if... for... will certainly fail)
    # do nothing for now
    return expr


def findUndeclaredVariables(ast, allUndeclaredVariables):
    """Detect undeclared variables
    """
    undeclaredVariables = []

    for expression in ast:
        if expression.type == 'ExpressionStatement':
            expression = expression.expression
            if expression.type == 'AssignmentExpression' and expression.operator == '=' and expression.left.type == 'Identifier':
                variableName = expression.left.name
                if variableName not in allUndeclaredVariables:
                    undeclaredVariables.append(variableName)
                    allUndeclaredVariables.append(variableName)
        elif expression.type == 'IfStatement':
            if expression.consequent.body is None:
                consequentVariables = findUndeclaredVariables(
                        [expression.consequent], allUndeclaredVariables)
            else:
                consequentVariables = findUndeclaredVariables(
                    expression.consequent.body, allUndeclaredVariables)
            undeclaredVariables.extend(consequentVariables)
        elif expression.type == "ReturnStatement":
            if expression.argument.type == "FunctionExpression":
                consequentVariables = findUndeclaredVariables(
                    expression.argument.body.body, allUndeclaredVariables)
                undeclaredVariables.extend(consequentVariables)
    return undeclaredVariables


def addVariableDeclarations(inputProgram, fileName):
    """Transform the input program by adding just before each function
    a declaration for its undeclared variables
    """

    # parse Javascript code into abstract syntax tree:
    # NB: esprima: https://media.readthedocs.org/pdf/esprima/4.0/esprima.pdf
    fileName = Path(str(fileName))
    try:
        ast = esprima.parseScript(inputProgram, {'range': True, 'tolerant': True})
    except esprima.error_handler.Error as err:
        if fileName:
            logging.error(f"Error parsing JS in {fileName.name}:\n{err}")
        else:
            logging.error(f"Error parsing JS: {err}")
        logging.flush()
        return inputProgram  # So JS can be written to file

    # find undeclared vars in functions and declare them before the function
    outputProgram = inputProgram
    offset = 0
    allUndeclaredVariables = []

    for expression in ast.body:
        if expression.type == 'FunctionDeclaration':
            # find all undeclared variables:
            undeclaredVariables = findUndeclaredVariables(expression.body.body,
                                                          allUndeclaredVariables)

            # add declarations (var) just before the function:
            funSpacing = ['', '\n'][len(undeclaredVariables) > 0]  # for consistent function spacing
            declaration = funSpacing + '\n'.join(['var ' + variable + ';' for variable in
                                     undeclaredVariables]) + '\n'
            startIndex = expression.range[0] + offset
            outputProgram = outputProgram[
                            :startIndex] + declaration + outputProgram[
                                                         startIndex:]
            offset += len(declaration)

    return outputProgram


if __name__ == '__main__':
    for expr in ['sin(t)', 't*5',
                 '(3, 4)', '(5*-2)',  # tuple and not tuple
                 '(1,(2,3), (1,2,3), (-4,-5,-6))', '2*(2, 3)',  # combinations
                 '[1, (2*2)]',  # List with nested operations returns list + nested tuple
                 '(.7, .7)',  # A tuple returns list
                 '(-.7, .7)',  # A tuple with unary operators returns nested lists
                 '[-.7, -.7]',  # A list with unary operators returns list with nested tuple
                 '[-.7, (-.7 * 7)]']:  # List with unary operators and nested tuple with operations returns list + tuple
        print("{} -> {}".format(repr(expr), repr(expression2js(expr))))



from psychopy import visual, event, info
import pytest
import numpy as np
import shutil, os
from tempfile import mkdtemp
from psychopy.tests import utils


# Testing for memory leaks in PsychoPy classes (experiment run-time, not Builder, Coder, etc)

# The tests are too unstable to include in travis-ci at this point.

# command-line usage:
# py.testw tests/test_misc/memory_usage.py

# Define the "acceptable" leakage severity; some gc vagaries are possible.
THRESHOLD = 0.5

win = visual.Window(size=(200,200), allowStencil=True)

def leakage(Cls, *args, **kwargs):
    """make up to 100 instances of Cls(*args, **kwargs),
    return the difference in memory used by this python process (in M) as a
    severity measure, approx = 100 * mem leak per instance in M
    """
    mem = []
    for i in range(100):
        Cls(*args, **kwargs)  # anonymous instance, gets gc'd each iteration
        mem.append(info.getMemoryUsage())
        # don't keep going if we're leaking:
        if mem[i] - mem[0] > THRESHOLD:
            break
    proportion = i / 99.
    return round((mem[i] - mem[0]) / proportion, 1)


@pytest.mark.needs_sound
@pytest.mark.memory
class TestMemorySound():
    @classmethod
    def setup_class(self):
        global sound, pyo
        from psychopy import sound
        import pyo
        self.tmp = mkdtemp(prefix='psychopy-tests-memory-usage')
    @classmethod
    def teardown_class(self):
        if hasattr(self, 'tmp'):
            shutil.rmtree(self.tmp, ignore_errors=True)

    def test_soundpyo_array(self):
        """anything using a numpy.array uses pyo.DataTable
        """
        if pyo.getVersion() < (0, 7, 7):
            pytest.xfail()  # pyo leak fixed Oct 2015
        for stim in [440, np.zeros(88200)]:  # np.zeros(8820000) passes, slow
            assert leakage(sound.SoundPyo, stim, secs=2) < THRESHOLD, 'stim = ' + str(stim)

    def test_soundpyo_file(self):
        """files are handled by pyo.SndFile
        """
        if pyo.getVersion() < (0, 7, 7):
            pytest.xfail()
        from scipy.io import wavfile
        tmp = os.path.join(self.tmp, 'zeros.wav')
        wavfile.write(tmp, 44100, np.zeros(88200))

        assert leakage(sound.SoundPyo, tmp) < THRESHOLD


@pytest.mark.needs_sound
@pytest.mark.memory
class TestMemoryMovie():
    @classmethod
    def setup_class(self):
        self.mov = os.path.join(utils.TESTS_DATA_PATH, 'testMovie.mp4')
    @classmethod
    def teardown_class(self):
        if hasattr(self, 'tmp'):
            shutil.rmtree(self.tmp, ignore_errors=True)

    def test_movie3_leakage(self):
        assert leakage(visual.MovieStim3, win, self.mov) < THRESHOLD

    @pytest.mark.skipif('True')
    def test_movie_leakage(self):
        assert leakage(visual.MovieStim, win, self.mov) < THRESHOLD

    @pytest.mark.skipif('True')
    def test_movie2_leakage(self):
        assert leakage(visual.MovieStim2, win, self.mov) < THRESHOLD


@pytest.mark.memory
class TestMemory():
    @classmethod
    def setup_class(self):
        self.imgs = [os.path.join(utils.TESTS_DATA_PATH, 'testimage.jpg'),  # smaller
                     os.path.join(utils.TESTS_DATA_PATH, 'greyscale.jpg')]  # larger
    @classmethod
    def teardown_class(self):
        if hasattr(self, 'tmp'):
            shutil.rmtree(self.tmp, ignore_errors=True)

    def test_Mouse(self):
        assert leakage(event.Mouse, win=win) < THRESHOLD

    def test_VisualStim(self):
        """Visual stim that typically do not leak can all be tested together
        """
        cleanStim = ['ShapeStim', 'Rect', 'Circle', 'Polygon', 'Line', 'CustomMouse', 'Aperture']
        for StimName in cleanStim:
            Stim = eval('visual.' + StimName)
            assert leakage(Stim, win) < THRESHOLD, StimName

    def test_ShapeStim(self):
        v = [(-.2,-.05), (-.2,.05), (.2,.05), (.2,.15), (.35,0), (.2,-.15), (.2,-.05)]
        assert leakage(visual.ShapeStim, win, vertices=v) < THRESHOLD
        assert leakage(visual.ShapeStim, win, vertices=v * 100) < THRESHOLD

    @pytest.mark.xfail
    def test_Window(self):
        msg = 'leakage probably not a problem for typical users with 1 Window() instance'
        assert leakage(visual.Window, size=(100, 100)) < THRESHOLD, msg
        assert leakage(visual.Window, size=(2000, 2000)) < THRESHOLD, msg

    def test_TextStim(self):
        msg = "Note: some TextStim leakage is pyglet's fault"
        for txt in ['a', 'a'*1000]:
            assert leakage(visual.TextStim, win, txt) < THRESHOLD, msg

    def test_RatingScale(self):
        msg = "RatingScale will probably leak if TextStim does"
        # 'hover' has few visual items (no text, line, marker, accept box)
        for kwargs in [{'marker': 'hover', 'choices': [1,2]}, {}]:
            assert leakage(visual.RatingScale, win, **kwargs) < THRESHOLD, msg

    def test_BufferImageStim(self):
        msg = "Note: the size of the window and the rect to capture affects leak severity"
        for rect in [(-.1,.1,.1,-.1), (-1,1,1,-1)]:
            assert leakage(visual.BufferImageStim, win, rect=rect) < THRESHOLD, msg

    def test_ImageStim(self):
        msg = "Note: the image size affects leak severity"
        for img in self.imgs:
            assert leakage(visual.ImageStim, win, img) < THRESHOLD, msg

    def test_SimpleImageStim(self):
        for img in self.imgs:
            assert leakage(visual.SimpleImageStim, win, img) < THRESHOLD

    def test_GratingStim(self):
        assert leakage(visual.GratingStim, win) < THRESHOLD

    def test_DotStim(self):
        assert leakage(visual.DotStim, win, nDots=2000) < THRESHOLD

    def test_RadialStim(self):
        for r in [4, 16]:
            assert leakage(visual.RadialStim, win, radialCycles=r, angularCycles=r) < THRESHOLD

    def test_ElementArrayStim(self):
        for n in [100, 1000]:
            assert leakage(visual.ElementArrayStim, win, nElements=n) < THRESHOLD



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Part of the PsychoPy library
# Copyright (C) 2002-2018 Jonathan Peirce (C) 2019-2022 Open Science Tools Ltd.
# Distributed under the terms of the GNU General Public License (GPL).

import numpy
import copy
from os import path
from psychopy import logging
from psychopy.constants import (STARTED, PLAYING, PAUSED, FINISHED, STOPPED,
                                NOT_STARTED, FOREVER)
from psychopy.tools.filetools import pathToString
from sys import platform
from .audioclip import AudioClip


if platform == 'win32':
    mediaLocation = "C:\\Windows\\Media"
elif platform == 'darwin':
    mediaLocation = "/System/Library/Sounds/"
elif platform.startswith("linux"):
    mediaLocation = "/usr/share/sounds"

stepsFromA = {
    'C': -9,
    'Csh': -8, 'C#': -8,
    'Dfl': -8, 'D♭': -8,
    'D': -7,
    'Dsh': -6, 'D#': -6,
    'Efl': -6, 'E♭': -6,
    'E': -5,
    'F': -4,
    'Fsh': -3, 'F#': -3,
    'Gfl': -3, 'G♭': -3,
    'G': -2,
    'Gsh': -1, 'G#': -1,
    'Afl': -1, 'A♭': -1,
    'A': +0,
    'Ash': +1, 'A#': +1,
    'Bfl': +1, 'B♭': +1,
    'B': +2,
    'Bsh': +2, 'B#': +2}
knownNoteNames = sorted(stepsFromA.keys())


def apodize(soundArray, sampleRate):
    """Apply a Hanning window (5ms) to reduce a sound's 'click' onset / offset
    """
    hwSize = int(min(sampleRate // 200, len(soundArray) // 15))
    hanningWindow = numpy.hanning(2 * hwSize + 1)
    soundArray = copy.copy(soundArray)
    soundArray[:hwSize] *= hanningWindow[:hwSize]
    soundArray[-hwSize:] *= hanningWindow[hwSize + 1:]
    return soundArray


class HammingWindow():
    def __init__(self, winSecs, soundSecs, sampleRate):
        """

        :param winSecs:
        :param soundSecs:
        :param sampleRate:
        """
        self.sampleRate = sampleRate
        self.winSecs = winSecs
        self.winSamples = int(round(sampleRate*winSecs))
        self.soundSecs = soundSecs
        self.soundSamples = int(round(sampleRate*soundSecs))
        self.startWindow = numpy.hanning(self.winSamples*2)[0:self.winSamples]
        self.endWindow = numpy.hanning(self.winSamples*2)[self.winSamples:]
        self.finalWinStart = self.soundSamples-self.winSamples

    def nextBlock(self, t, blockSize):
        """Returns a block to be multiplied with the current sound block or 1.0

        :param t: current position in time (secs)
        :param blockSize: block size for the sound needing the hanning window
        :return: numpy array of length blockSize
        """
        startSample = int(t*self.sampleRate)
        if startSample < self.winSamples:
            # we're in beginning hanning window (start of sound)
            # 2 options:
            #  - block is fully within window
            #  - block starts in window but ends after window
            block = numpy.ones(blockSize)
            winEndII = min(self.winSamples,  # if block goes beyond hann win
                           startSample+blockSize)  # if block shorter
            blockEndII = min(self.winSamples-startSample,  # if block beyond
                             blockSize)  # if block shorter
            block[0:blockEndII] = self.startWindow[startSample:winEndII]
        elif startSample >= self.soundSamples:
            block = None  # the sound has finished (shouldn't have got here!)
        elif startSample >= self.finalWinStart-blockSize:
            # we're in final hanning window (end of sound)
            # More complicated, with 3 options:
            #  - block starts before win
            #  - start/end during win
            #  - start during but end after win
            block = numpy.ones(blockSize)  # the initial flat part
            blockStartII = max(self.finalWinStart-startSample,
                         0)  # if block start inside window
            blockEndII = min(blockSize,  # if block ends in hann win
                             self.soundSamples-startSample)  # ends after snd
            winStartII = max(0,  # current block ends in win but starts before
                             startSample-self.finalWinStart)
            winEndII = min(self.winSamples,
                           startSample+blockSize-self.finalWinStart)
            block[blockStartII:blockEndII] = \
                self.endWindow[winStartII:winEndII]
        else:
            block = None  # we're in the middle of sound so no need for window
        if block is not None:
            block.shape = [len(block), 1]
        return block


class _SoundBase():
    """Base class for sound object, from one of many ways.
    """
    # Must be provided by class SoundPygame or SoundPyo:
    # def __init__()
    # def play(self, fromStart=True, **kwargs):
    # def stop(self, log=True):
    # def getDuration(self):
    # def getVolume(self):
    # def setVolume(self, newVol, log=True):
    # def _setSndFromFile(self, fileName):
    # def _setSndFromArray(self, thisArray):

    def setSound(self, value, secs=0.5, octave=4, hamming=True, log=True):
        """Set the sound to be played.

        Often this is not needed by the user - it is called implicitly during
        initialisation.

        Parameters
        ----------
        value : ArrayLike, AudioClip, int or str
            If it's a number between 37 and 32767 then a tone will be generated
            at that frequency in Hz. It could be a string for a note ('A',
            'Bfl', 'B', 'C', 'Csh'. ...). Then you may want to specify which
            octave.O r a string could represent a filename in the current
            location, or media location, or a full path combo. Or by giving an
            Nx2 numpy array of floats (-1:1) or
            :class:`~psychopy.sound.AudioClip` instance.
        secs : float
            Duration of a tone if a note name is to `value`.
        octave : int
            Is only relevant if the value is a note name. Middle octave of a
            piano is 4. Most computers won't output sounds in the bottom octave
            (1) and the top octave (8) is generally painful.
        hamming : bool
            To indicate if the sound should be apodized (i.e., the onset and
            offset smoothly ramped up from down to zero). The function apodize
            uses a Hanning window, but arguments named 'hamming' are preserved
            so that existing code is not broken by the change from Hamming to
            Hanning internally. Not applied to sounds from files.

        """
        # Re-init sound to ensure bad values will raise error during setting:
        self._snd = None

        # Coerces pathlib obj to string, else returns inputted value
        value = pathToString(value)
        try:
            # could be '440' meaning 440
            value = float(value)
        except (ValueError, TypeError):
            # this is a string that can't be a number, eg, a file or note
            pass
        else:
            # we've been asked for a particular Hz
            if value < 37 or value > 20000:
                msg = 'Sound: bad requested frequency %.0f'
                raise ValueError(msg % value)
            self._setSndFromFreq(value, secs, hamming=hamming)
        if isinstance(value, str):
            if value.capitalize() in knownNoteNames:
                self._setSndFromNote(value.capitalize(), secs, octave,
                                     hamming=hamming)
            else:
                # try finding a file
                self.fileName = None
                for filePath in ['', mediaLocation]:
                    p = path.join(filePath, value)
                    if path.isfile(p):
                        self.fileName = p
                        break
                    elif path.isfile(p + '.wav'):
                        self.fileName = p = p + '.wav'
                        break
                if self.fileName is None:
                    msg = "setSound: could not find a sound file named "
                    raise ValueError(msg + value)
                else:
                    self._setSndFromFile(p)
        elif isinstance(value, (list, numpy.ndarray,)):
            # create a sound from the input array/list
            self._setSndFromArray(numpy.array(value))
        elif isinstance(value, AudioClip):
            # from an audio clip object
            self._setSndFromArray(value.samples)
            self.sampleRate = value.sampleRateHz
        # did we succeed?
        if self._snd is None:
            pass  # raise ValueError, "Could not make a "+value+" sound"
        else:
            if log and self.autoLog:
                logging.exp("Set %s sound=%s" % (self.name, value), obj=self)
            self.status = NOT_STARTED

    def _setSndFromNote(self, thisNote, secs, octave, hamming=True):
        # note name -> freq -> sound
        freqA = 440.0
        thisOctave = octave - 4
        mult = 2.0**(stepsFromA[thisNote] / 12.)
        thisFreq = freqA * mult * 2.0 ** thisOctave
        self._setSndFromFreq(thisFreq, secs, hamming=hamming)

    def _setSndFromFreq(self, thisFreq, secs, hamming=True):
        # note freq -> array -> sound
        if secs < 0:
            # want infinite duration - create 1 sec sound and loop it
            secs = 10.0
            self.loops = -1
        if not self.sampleRate:
            self.sampleRate = self._getDefaultSampleRate()
        nSamples = int(secs * self.sampleRate)
        outArr = numpy.arange(0.0, 1.0, 1.0 / nSamples)
        outArr *= 2 * numpy.pi * thisFreq * secs
        outArr = numpy.sin(outArr)
        if hamming and nSamples > 30:
            outArr = apodize(outArr, self.sampleRate)
        self._setSndFromArray(outArr)

    def _getDefaultSampleRate(self):
        """For backends this might depend on what streams are open"""
        return 44100

    def getDuration(self):
        """Return the duration of the sound"""
        return self.duration

    def getVolume(self):
        """Returns the current volume of the sound (0.0 to 1.0, inclusive)
        """
        return self.volume

    def getLoops(self):
        """Returns the current requested loops value for the sound (int)
        """
        return self.loops

    def setVolume(self, newVol, log=True):
        """Sets the current volume of the sound (0.0 to 1.0, inclusive)
        """
        self.volume = min(1.0, max(0.0, newVol))
        self.needsUpdate = True
        if log and self.autoLog:
            logging.exp("Sound %s set volume %.3f" %
                        (self.name, self.volume), obj=self)

    def setLoops(self, newLoops, log=True):
        """Sets the current requested extra loops (int)"""
        self.loops = int(newLoops)
        self.needsUpdate = True
        if log and self.autoLog:
            logging.exp("Sound %s set loops %s" %
                        (self.name, self.loops), obj=self)



import sys
from pyglet.window import key
from psychopy.visual import Window, TextStim, GratingStim, Circle
from psychopy.visual.windowwarp import Warper
from psychopy import event
import pytest, copy

"""define WindowWarp configurations, test the logic

    test:
    cd psychopy/psychopy/
    py.test -k projections --cov-report term-missing --cov visual/windowwarp.py
"""

foregroundColor = [-1, -1, -1]
backgroundColor = [1, 1, 1]


class ProjectionsLinesAndCircles:
    """
    Test jig for projection warping.
    Switch between warpings by pressing a key 'S'pherical, 'C'ylindrical, 'N'one, warp'F'ile.
    Click the mouse to set the eyepoint X, Y.
    Up / Down arrow or mousewheel to move eyepoint in and out.
    """
    def __init__(self, win, warper):

        self.win = win
        self.warper = warper

        self.stimT = TextStim(self.win, text='Null warper', units = 'pix',
                              pos=(0, -140), height=20)

        self.bl = -win.size / 2.0
        self.tl = (self.bl[0], -self.bl[1])
        self.tr = win.size / 2.0

        self.stims = []
        self.degrees = 120
        nLines = 12
        for x in range(-nLines, nLines+1):
            t = GratingStim(win, tex=None, units='deg', size=[2, win.size[1]],
                            texRes=128, color=foregroundColor,
                            pos=[float(x) / nLines * self.degrees, 0])
            self.stims.append (t)

        for y in range (-nLines, nLines+1):
            t = GratingStim(win, tex=None, units='deg', size=[win.size[0], 2],
                            texRes=128, color=foregroundColor,
                            pos=[0, float(y)/nLines * self.degrees])
            self.stims.append(t)

        for c in range (1, nLines+1):
            t = Circle (win, radius=c * 10, edges=128, units='deg', lineWidth=4)
            self.stims.append(t)

        self.updateInfo()

        self.keys = key.KeyStateHandler()
        win.winHandle.push_handlers(self.keys)
        self.mouse = event.Mouse(win=self.win)

    def updateFrame(self):
        """ Updates frame with any item that is to be modulated per frame. """
        for s in self.stims:
            s.draw()
        self.stimT.draw()

    def update_sweep(self):
        """ Update function for sweeps. Input is in domain units. """
        self.updateFrame()
        self.check_keys()
        self._handleMouse()
        self.win.flip()

    def updateInfo(self):
        try:
            self.stimT.setText(
                "%s \n   eyePoint: %.3f, %.3f \n   eyeDistance: %.2f\n\n"
                "Projection: [s]pherical, [c]ylindrical, [n]one, warp[f]ile\n"
                "Flip: [h]orizontal, [v]ertical\n"
                "Mouse: wheel = eye distance, click to set eyepoint\n"
                "[q]uit" % (
                self.warper.warp,
                self.warper.eyepoint[0], self.warper.eyepoint[1],
                self.warper.dist_cm))
        except Exception:
            pass

    def check_keys(self):
        """Checks key input"""
        for keys in event.getKeys(timeStamped=True):
            k = keys[0]
            if k in ['escape', 'q']:
                self.win.close()
                sys.exit()
            elif k in ['space']:
                for c in range (1,2):
                    t = Circle(self.win, radius=c)
                    self.stims.append (t)
                #for c in range (1,2):
                #    t = RadialStim(self.win)
                #    self.stims.append(t)

            # handle projections
            elif k in ['s']:
                self.warper.changeProjection ('spherical', None, (0.5,0.5))
            elif k in ['c']:
                self.warper.changeProjection ('cylindrical', None, (0.5,0.5))
            elif k in ['n']:
                self.warper.changeProjection (None, None, (0.5,0.5))
            elif k in ['f']:
                self.warper.changeProjection ('warpfile',
                    r'..\data\sample.meshwarp.data',
                     (0.5,0.5))

            # flip horizontal and vertical
            elif k in ['h']:
                self.warper.changeProjection(self.warper.warp, self.warper.warpfile, flipHorizontal = not self.warper.flipHorizontal)
            elif k in ['v']:
                self.warper.changeProjection(self.warper.warp, self.warper.warpfile, flipVertical = not self.warper.flipVertical)

            # move eyepoint
            elif k in ['down']:
                if (self.warper.dist_cm > 1):
                    self.warper.dist_cm -= 1
                    self.warper.changeProjection (self.warper.warp, None, self.warper.eyepoint)
            elif k in ['up']:
                if (self.warper.dist_cm < 200):
                    self.warper.dist_cm += 1
                    self.warper.changeProjection (self.warper.warp, None, self.warper.eyepoint)
            elif k in ['right']:
                if (self.warper.eyepoint[0] < 0.9):
                    self.warper.eyepoint = (self.warper.eyepoint[0] + 0.1, self.warper.eyepoint[1])
                    self.warper.changeProjection (self.warper.warp, None, self.warper.eyepoint)
            elif k in ['left']:
                if (self.warper.eyepoint[0] > 0.1):
                    self.warper.eyepoint = (self.warper.eyepoint[0] - 0.1, self.warper.eyepoint[1])
                    self.warper.changeProjection (self.warper.warp, None, self.warper.eyepoint)

            self.updateInfo()

    def _handleMouse(self):
        x,y = self.mouse.getWheelRel()
        if y != 0:
            self.warper.dist_cm += y
            self.warper.dist_cm = max (1, min (200, self.warper.dist_cm))
            self.warper.changeProjection (self.warper.warp, self.warper.warpfile, self.warper.eyepoint)
            self.updateInfo()

        pos = (self.mouse.getPos() + 1) / 2
        leftDown = self.mouse.getPressed()[0]
        if leftDown:
            self.warper.changeProjection (self.warper.warp, self.warper.warpfile, pos)
            self.updateInfo()


class Test_class_WindowWarp():
    def setup_class(self):
        self.win = Window(monitor='testMonitor', screen=1, fullscr=True, color='gray', useFBO = True, autoLog=False)
        self.warper = Warper(self.win, warp='spherical', warpfile="", warpGridsize=128, eyepoint=[0.5, 0.5],
                             flipHorizontal=False, flipVertical=False)
        self.warper.dist_cm = 15
        self.g = ProjectionsLinesAndCircles(self.win, self.warper)

    def teardown_class(self):
        self.win.close()

    def draw_projection (self, frames=120):
        self.g.updateInfo()
        for i in range(frames):
            self.g.update_sweep()

    def test_spherical(self):
        self.warper.changeProjection('spherical')
        self.draw_projection()

    def test_cylindrical(self):
        self.warper.changeProjection('cylindrical')
        self.draw_projection()

    def test_warpfile(self):
        self.warper.changeProjection('warpfile', warpfile="") #jayb todo
        self.draw_projection()

    def test_distance(self):
        self.test_spherical()
        for i in range (1, 50, 2):
            self.warper.dist_cm = i
            self.warper.changeProjection(self.warper.warp)
            self.g.updateInfo()
            self.g.update_sweep()

        self.test_cylindrical()
        for i in range (1, 50, 2):
            self.warper.dist_cm = i
            self.warper.changeProjection(self.warper.warp)
            self.g.updateInfo()
            self.g.update_sweep()

    def test_flipHorizontal(self):
        self.warper.changeProjection(self.warper.warp, self.warper.warpfile, flipHorizontal = not self.warper.flipHorizontal)
        self.draw_projection()

    def test_flipVertical(self):
        self.warper.changeProjection(self.warper.warp, self.warper.warpfile, flipVertical = not self.warper.flipVertical)
        self.draw_projection()

if __name__ == '__main__':
        # running interactive
        cls = Test_class_WindowWarp()
        cls.setup_class()
        for i in range(2 * 60 * 60):
            cls.g.update_sweep()
        cls.win.close()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Demo: show a very basic program: hello world
"""

# Import key parts of the PsychoPy library:
from psychopy import visual, core

# Create a visual window:
win = visual.Window(units="height")

# Create (but not yet display) some text:
msg1 = visual.TextBox2(win, 
    text=u"Hello world!", 
    font="Open Sans", letterHeight=0.1,
    pos=(0, 0.2)) 
msg2 = visual.TextBox2(win, 
    text=u"\u00A1Hola mundo!", 
    font="Open Sans", letterHeight=0.1, 
    pos=(0, -0.2))

# Draw the text to the hidden visual buffer:
msg1.draw()
msg2.draw()

# Show the hidden buffer--everything that has been drawn since the last win.flip():
win.flip()

# Wait 3 seconds so people can see the message, then exit gracefully:
core.wait(3)

win.close()
core.quit()

# The contents of this file are in the public domain.



# -*- coding: utf-8 -*-
# Part of the PsychoPy library
# Copyright (C) 2012-2020 iSolver Software Solutions (C) 2021 Open Science Tools Ltd.
# Distributed under the terms of the GNU General Public License (GPL).
from .. import Device, ioDeviceError
from ...constants import DeviceConstants, EyeTrackerConstants
from . import hw
from ...errors import print2err


class EyeTrackerDevice(Device):
    """The EyeTrackerDevice class is the main class for the ioHub Common Eye
    Tracker interface.

    The Common Eye Tracker Interface--a set of common functions and methods
    such that the same experiment script and data analyses can be shared,
    used, and compared regardless of the actual eye tracker used--works by
    extending the EyeTrackerDevice class to configure device monitoring and
    data access to individual eye tracker manufacturers and models.

    Not every EyeTrackerDevice subclass will support all of the umbrella functionality
    within the Common Eye Tracker Interface, but a core set of critical functions are
    supported by all eye tracker models to date. Any Common Eye Tracker Interface
    method not supported by the configured Eye Tracker hardware returns a constant
    (EyeTrackerConstants.FUNCTIONALITY_NOT_SUPPORTED).

    Methods in the EyeTrackerDevice class are broken down into several categories:

    #. Initializing the Eye Tracker / Setting the Device State.
    #. Defining the Graphics Layer for Calibration / System Setup.
    #. Starting and Stopping of Data Recording.
    #. Sending Messages or Codes to Synchronize the ioHub with the Eye Tracker.
    #. Accessing Eye Tracker Data During Recording.
    #. Accessing the Eye Tracker native time base.
    #. Synchronizing the ioHub time base with the Eye Tracker time base

    .. note::

        Only **one** instance of EyeTracker can be created within an experiment.
        Attempting to create > 1 instance will raise an exception.

    """

    # Used to hold the EyeTracker subclass instance to ensure only one instance of
    # a given eye tracker type is created. This is a current ioHub limitation, not the limitation of
    # all eye tracking hardware.
    _INSTANCE = None

    #: The multiplier needed to convert a device's native time base to sec.msec-usec times.
    DEVICE_TIMEBASE_TO_SEC = 1.0

    # Used by pyEyeTrackerDevice implementations to store relationships between an eye
    # trackers command names supported for EyeTrackerDevice sendCommand method and
    # a private python function to call for that command. This allows an implementation
    # of the interface to expose functions that are not in the core EyeTrackerDevice spec
    # without have to use the EXT extension class.
    _COMMAND_TO_FUNCTION = {}

    DEVICE_TYPE_ID = DeviceConstants.EYETRACKER
    DEVICE_TYPE_STRING = 'EYETRACKER'
    __slots__ = [
        '_latest_sample',
        '_latest_gaze_position',
        '_runtime_settings']

    def __init__(self, *args, **kwargs):
        if self.__class__._INSTANCE is not None:
            raise ioDeviceError(
                self, 'EyeTracker object has already been created; '
                'only one instance can exist. Delete existing '
                'instance before recreating EyeTracker object.')
        else:
            self.__class__._INSTANCE = self

        Device.__init__(self, *args, **kwargs['dconfig'])

        # hold last received ioHub eye sample (in ordered array format) from
        # tracker.
        self._latest_sample = EyeTrackerConstants.FUNCTIONALITY_NOT_SUPPORTED

        # holds the last gaze position read from the eye tracker as an x,y tuple. If binocular recording is
        # being performed, this is an average of the left and right gaze
        # position x,y fields.
        self._latest_gaze_position = EyeTrackerConstants.FUNCTIONALITY_NOT_SUPPORTED

        # stores the eye tracker runtime related configuration settings from
        # the ioHub .yaml config file
        self._runtime_settings = kwargs[
            'dconfig'].get('runtime_settings', None)

        # TODO: Add support for message ID to Message text lookup table in ioDataStore
        # data table that can be used by ET systems that support sending int codes,
        # but not text to tracker at runtime for syncing.

    def trackerTime(self):
        """trackerTime returns the current time reported by the eye tracker
        device. The time base is implementation dependent.

        Args:
            None

        Return:
            float: The eye tracker hardware's reported current time.

        """
        return EyeTrackerConstants.FUNCTIONALITY_NOT_SUPPORTED

    def trackerSec(self):
        """
        trackerSec takes the time received by the EyeTracker.trackerTime() method
        and returns the time in sec.usec-msec format.

        Args:
            None

        Return:
            float: The eye tracker hardware's reported current time in sec.msec-usec format.
        """
        return EyeTrackerConstants.FUNCTIONALITY_NOT_SUPPORTED

    def setConnectionState(self, enable):
        """setConnectionState either connects ( setConnectionState(True) ) or
        disables ( setConnectionState(False) ) active communication between the
        ioHub and the Eye Tracker.

        .. note::
            A connection to the Eye Tracker is automatically established
            when the ioHub Process is initialized (based on the device settings
            in the iohub_config.yaml), so there is no need to
            explicitly call this method in the experiment script.

        .. note::
            Connecting an Eye Tracker to the ioHub does **not** necessarily collect and send
            eye sample data to the ioHub Process. To start actual data collection,
            use the Eye Tracker method setRecordingState(bool) or the ioHub Device method (device type
            independent) enableEventRecording(bool).

        Args:
            enable (bool): True = enable the connection, False = disable the connection.

        Return:
            bool: indicates the current connection state to the eye tracking hardware.

        """
        return EyeTrackerConstants.FUNCTIONALITY_NOT_SUPPORTED

    def isConnected(self):
        """isConnected returns whether the ioHub EyeTracker Device is connected
        to the eye tracker hardware or not. An eye tracker must be connected to
        the ioHub for any of the Common Eye Tracker Interface functionality to
        work.

        Args:
            None

        Return:
            bool:  True = the eye tracking hardware is connected. False otherwise.

        """
        return EyeTrackerConstants.FUNCTIONALITY_NOT_SUPPORTED

    def sendCommand(self, key, value=None):
        """
        The sendCommand method allows arbitrary *commands* or *requests* to be
        issued to the eye tracker device. Valid values for the arguments of this
        method are completely implementation-specific, so please refer to the
        eye tracker implementation page for the eye tracker being used for a list of
        valid key and value combinations (if any).

        In general, eye tracker implementations should **not** need to support
        this method unless there is critical eye tracker functionality that is
        not accessible using the other methods in the EyeTrackerDevice class.

        Args:
            key (str): the command or function name that should be run.
            value (object): the (optional) value associated with the key.

        Return:
            object: the result of the command call
            int: EyeTrackerConstants.EYETRACKER_OK
            int: EyeTrackerConstants.EYETRACKER_ERROR
            int: EyeTrackerConstants.EYETRACKER_INTERFACE_METHOD_NOT_SUPPORTED
        """

        return EyeTrackerConstants.FUNCTIONALITY_NOT_SUPPORTED

    def sendMessage(self, message_contents, time_offset=None):
        """The sendMessage method sends a text message to the eye tracker.

        Messages are generally used to send information you want
        saved with the native eye data file and are often used to
        synchronize stimulus changes in the experiment with the eye
        data stream being saved to the native eye tracker data file (if any).

        This means that the sendMessage implementation needs to
        perform in real-time, with a delay of <1 msec from when a message is
        sent to when it is time stamped by the eye tracker, for it to be
        accurate in this regard.

        If this standard can not be met, the expected delay and message
        timing precision (variability) should be provided in the eye tracker's
        implementation notes.

        .. note::
            If using the ioDataStore to save the eye tracker data, the use of
            this method is quite optional, as Experiment Device Message Events
            will likely be preferred. ioHub Message Events are stored in the ioDataStore,
            alongside all other device data collected via the ioHub, and not
            in the native eye tracker data.

        Args:
           message_contents (str):
               If message_contents is a string, check with the implementations documentation if there are any string length limits.

        Kwargs:
           time_offset (float): sec.msec_usec time offset that the time stamp of
                              the message should be offset in the eye tracker data file.
                              time_offset can be used so that a message can be sent
                              for a display change **BEFORE** or **AFTER** the actual
                              flip occurred, using the following formula:

                              time_offset = sendMessage_call_time - event_time_message_represent

                              Both times should be based on the iohub.devices.Computer.getTime() time base.

                              If time_offset is not supported by the eye tracker implementation being used, a warning message will be printed to stdout.

        Return:
            (int): EyeTrackerConstants.EYETRACKER_OK, EyeTrackerConstants.EYETRACKER_ERROR, or EyeTrackerConstants.EYETRACKER_INTERFACE_METHOD_NOT_SUPPORTED

        """

        return EyeTrackerConstants.FUNCTIONALITY_NOT_SUPPORTED

    def runSetupProcedure(self, calibration_args={}):
        """
        The runSetupProcedure method starts the eye tracker calibration
        routine. If calibration_args are provided, they should be used to
        update calibration related settings prior to starting the calibration.
        
        The details of this method are implementation-specific.

        .. note::
            This is a blocking call for the PsychoPy Process
            and will not return to the experiment script until the necessary steps
            have been completed so that the eye tracker is ready to start collecting
            eye sample data when the method returns.

        Args:
            None
        """
        return EyeTrackerConstants.EYETRACKER_INTERFACE_METHOD_NOT_SUPPORTED

    def setRecordingState(self, recording):
        """The setRecordingState method is used to start or stop the recording
        and transmission of eye data from the eye tracking device to the ioHub
        Process.

        Args:
            recording (bool): if True, the eye tracker will start recordng data.; false = stop recording data.

        Return:
            bool: the current recording state of the eye tracking device

        """
        # Implementation Note: Perform your implementation specific logic for
        # this method here
        print2err(
            'EyeTracker should handle setRecordingState method with recording value of {0} now.'.format(recording))

        # Implementation Note: change current_recording_state to be True or
        # False, based on whether the eye tracker is now recording or not.
        current_recording_state = EyeTrackerConstants.EYETRACKER_INTERFACE_METHOD_NOT_SUPPORTED
        return current_recording_state

    def isRecordingEnabled(self):
        """The isRecordingEnabled method indicates if the eye tracker device is
        currently recording data.

        Args:
           None

        Return:
            bool: True == the device is recording data; False == Recording is not occurring

        """

        # Implementation Note: Perform your implementation specific logic for
        # this method here
        print2err('EyeTracker should handle isRecordingEnabled method now.')

        # Implementation Note: change is_recording to be True or False, based
        # on whether the eye tracker is recording or not.
        is_recording = EyeTrackerConstants.EYETRACKER_INTERFACE_METHOD_NOT_SUPPORTED

        return is_recording

    def getLastSample(self):
        """The getLastSample method returns the most recent eye sample received
        from the Eye Tracker. The Eye Tracker must be in a recording state for
        a sample event to be returned, otherwise None is returned.

        Args:
            None

        Returns:
            int: If this method is not supported by the eye tracker interface, EyeTrackerConstants.FUNCTIONALITY_NOT_SUPPORTED is returned.

            None: If the eye tracker is not currently recording data.

            EyeSample: If the eye tracker is recording in a monocular tracking mode, the latest sample event of this event type is returned.

            BinocularEyeSample:  If the eye tracker is recording in a binocular tracking mode, the latest sample event of this event type is returned.

        """

        return self._latest_sample

    def getLastGazePosition(self):
        """The getLastGazePosition method returns the most recent eye gaze
        position received from the Eye Tracker. This is the position on the
        calibrated 2D surface that the eye tracker is reporting as the current
        eye position. The units are in the units in use by the ioHub Display
        device.

        If binocular recording is being performed, the average position of both
        eyes is returned.

        If no samples have been received from the eye tracker, or the
        eye tracker is not currently recording data, None is returned.

        Args:
            None

        Returns:
            int: If this method is not supported by the eye tracker interface, EyeTrackerConstants.EYETRACKER_INTERFACE_METHOD_NOT_SUPPORTED is returned.

            None: If the eye tracker is not currently recording data or no eye samples have been received.

            tuple: Latest (gaze_x,gaze_y) position of the eye(s)
        """
        return self._latest_gaze_position

    def getPosition(self):
        """
        See getLastGazePosition().
        """
        return self.getLastGazePosition()

    def getPos(self):
        """
        See getLastGazePosition().
        """
        return self.getLastGazePosition()

    def _eyeTrackerToDisplayCoords(self, eyetracker_point):
        """The _eyeTrackerToDisplayCoords method is required for implementation
        of the Common Eye Tracker Interface in order to convert the native Eye
        Tracker coordinate space to the ioHub.devices.Display coordinate space
        being used in the PsychoPy experiment. Any screen based coordinates
        that exist in the data provided to the ioHub by the device
        implementation must use this method to convert the x,y eye tracker
        point to the correct coordinate space.

        Default implementation is to call the Display device method:

            self._display_device._pixel2DisplayCoord(gaze_x,gaze_y,self._display_device.getIndex())

        where gaze_x,gaze_y = eyetracker_point, which is assumed to be in screen pixel
        coordinates, with a top-left origin. If the eye tracker provides the eye position
        data in a coordinate space other than screen pixel position with top-left origin,
        the eye tracker position should first be converted to this coordinate space before
        passing the position data px,py to the _pixel2DisplayCoord method.

        self._display_device.getIndex() provides the index of the display for multi display setups.
        0 is the default index, and valid values are 0 - N-1, where N is the number
        of connected, active, displays on the computer being used.

        Args:
            eyetracker_point (object): eye tracker implementation specific data type representing an x, y position on the calibrated 2D plane (typically a computer display screen).

        Returns:
            (x,y): The x,y eye position on the calibrated surface in the current ioHub.devices.Display coordinate type and space.

        """
        gaze_x = eyetracker_point[0]
        gaze_y = eyetracker_point[1]

        # If the eyetracker_point does not represent eye data as display
        # pixel position using a top-left origin, convert the naive eye tracker
        # gaze coordinate space to a display pixel position using a top-left origin
        # here before passing gaze_x,gaze_y to the _pixel2DisplayCoord method.
        # ....

        return self._display_device._pixel2DisplayCoord(
            gaze_x, gaze_y, self._display_device.getIndex())

    def _displayToEyeTrackerCoords(self, display_x, display_y):
        """The _displayToEyeTrackerCoords method must be used by an eye
        trackers implementation of the Common Eye Tracker Interface to convert
        any gaze positions provided by the ioHub to the appropriate x,y gaze
        position coordinate space for the eye tracking device in use.

        This method is simply the inverse operation performed by the _eyeTrackerToDisplayCoords
        method.

        Default implementation is to just return the result of self._display_device.display2PixelCoord(...).

        Args:
            display_x (float): The horizontal eye position on the calibrated 2D surface in ioHub.devices.Display coordinate space.
            display_y (float): The vertical eye position on the calibrated 2D surface in ioHub.devices.Display coordinate space.

        Returns:
            (object): eye tracker implementation specific data type representing an x, y position on the calibrated 2D plane (typically a computer display screen).

        """

        pixel_x, pixel_y = self._display_device.display2PIxelCoord(
            display_x, display_y, self._display_device.getIndex())
        return pixel_x, pixel_y

    def __del__(self):
        """Do any final cleanup of the eye tracker before the object is
        destroyed."""
        self.__class__._INSTANCE = None

from .eye_events import (
    EyeSampleEvent,
    MonocularEyeSampleEvent,
    BinocularEyeSampleEvent,
    FixationStartEvent,
    FixationEndEvent,
    SaccadeStartEvent,
    SaccadeEndEvent,
    BlinkStartEvent,
    BlinkEndEvent)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Example of how to access Keyboard events using iohub.
Displays information from ioHub Keyboard Events vs. psychopy.event.geKeys().
"""

from __future__ import absolute_import, division, print_function

from builtins import str
from psychopy import core, visual, event
from psychopy.hardware import keyboard as ptb_keyboard
from psychopy.iohub import launchHubServer

WINDOW_SIZE = 1024, 768
dw = WINDOW_SIZE[0] / 2
dh = WINDOW_SIZE[1] / 2
unit_type = 'pix'
win = visual.Window(WINDOW_SIZE, units=unit_type,
    color=[128, 128, 128], colorSpace='rgb255')

# Create a Keyboard class with ptb as the backend
ptb_keyboard = ptb_keyboard.Keyboard(backend='ptb')

# Start iohub process. The iohub process can be accessed using `io`.
io = launchHubServer(window=win, Keyboard=dict(use_keymap='psychopy'))

# A `keyboard` variable is used to access the iohub Keyboard device.
keyboard = io.devices.keyboard


# constants for text element spacing:
ROW_COUNT = 10
TEXT_ROW_HEIGHT = (dh * 2) / ROW_COUNT
TEXT_STIM_HEIGHT = int(TEXT_ROW_HEIGHT / 2)
MARGIN = 25
LABEL_COLUMN_X = -dw + MARGIN
VALUE_COLUMN_X = MARGIN
LABEL_WRAP_LENGTH = dw - MARGIN / 2
VALUE_WRAP_LENGTH = dw - MARGIN / 2
TEXT_ROWS_START_Y = dh - MARGIN

# Create some psychoPy stim to display the keyboard events received...

# field labels:
title_label = visual.TextStim(win, units=unit_type,
    text=u'Press and Releases Keys for ioHub KB Event Details',
    pos=[0, TEXT_ROWS_START_Y],
    height=TEXT_STIM_HEIGHT,
    color='black', wrapWidth=dw * 2)
title2_label = visual.TextStim(win, units=unit_type,
    text=u'Press "Q" Key to Exit Demo',
    pos=[0, TEXT_ROWS_START_Y - TEXT_ROW_HEIGHT],
    height=TEXT_STIM_HEIGHT,
    color=[0.25, 0.2, 1],
    alignText='center', anchorHoriz='center', anchorVert='top',
    wrapWidth=dw * 2)
key_text_label = visual.TextStim(win, units=unit_type, text=u'iohub .key:',
    pos=[LABEL_COLUMN_X, TEXT_ROWS_START_Y - TEXT_ROW_HEIGHT * 2],
    color='black', alignText='left', anchorHoriz='left',
    height=TEXT_STIM_HEIGHT, wrapWidth=LABEL_WRAP_LENGTH)
char_label = visual.TextStim(win, units=unit_type, text=u'iohub .char:',
    pos=[LABEL_COLUMN_X, TEXT_ROWS_START_Y - TEXT_ROW_HEIGHT * 3],
    color='black', alignText='left', anchorHoriz='left',
    height=TEXT_STIM_HEIGHT, wrapWidth=LABEL_WRAP_LENGTH)
modifiers_label = visual.TextStim(win, units=unit_type,
    text=u'iohub .modifiers',
    pos=[LABEL_COLUMN_X, TEXT_ROWS_START_Y - TEXT_ROW_HEIGHT * 4],
    color='black', alignText='left', anchorHoriz='left',
    height=TEXT_STIM_HEIGHT, wrapWidth=LABEL_WRAP_LENGTH)
keypress_duration_label = visual.TextStim(win, units=unit_type,
    text=u'Last Pressed Duration:',
    pos=[LABEL_COLUMN_X, TEXT_ROWS_START_Y - TEXT_ROW_HEIGHT * 5],
    color='black', alignText='left', anchorHoriz='left',
    height=TEXT_STIM_HEIGHT, wrapWidth=LABEL_WRAP_LENGTH)
all_pressed__label = visual.TextStim(win, units=unit_type,
    text=u'All Pressed Keys:',
    pos=[LABEL_COLUMN_X, TEXT_ROWS_START_Y - TEXT_ROW_HEIGHT * 6],
    color='black', alignText='left', anchorHoriz='left',
    height=TEXT_STIM_HEIGHT, wrapWidth=LABEL_WRAP_LENGTH)
event_type_label = visual.TextStim(win, units=unit_type,
    text=u'Last Event Type:',
    pos=[LABEL_COLUMN_X, TEXT_ROWS_START_Y - TEXT_ROW_HEIGHT * 7],
    color='black', alignText='left', anchorHoriz='left',
    height=TEXT_STIM_HEIGHT, wrapWidth=LABEL_WRAP_LENGTH)
psychopy_key_label = visual.TextStim(win, units=unit_type,
    text=u'vs. event.getKeys():',
    pos=[LABEL_COLUMN_X, TEXT_ROWS_START_Y - TEXT_ROW_HEIGHT * 8],
    color='black', alignText='left', anchorHoriz='left',
    height=TEXT_STIM_HEIGHT, wrapWidth=LABEL_WRAP_LENGTH)
ptb_key_label = visual.TextStim(win, units=unit_type,
    text=u'vs. ptb_kb.getKeys():',
    pos=[LABEL_COLUMN_X, TEXT_ROWS_START_Y - TEXT_ROW_HEIGHT * 9],
    color='black', alignText='left', anchorHoriz='left',
    height=TEXT_STIM_HEIGHT, wrapWidth=LABEL_WRAP_LENGTH)

# Dynamic stim:
key_text_stim = visual.TextStim(win, units=unit_type, text=u'',
    pos=[VALUE_COLUMN_X, TEXT_ROWS_START_Y - TEXT_ROW_HEIGHT * 2],
    color='black', alignText='left', anchorHoriz='left',
    height=TEXT_STIM_HEIGHT, wrapWidth=LABEL_WRAP_LENGTH)
char_stim = visual.TextStim(win, units=unit_type, text=u'',
    pos=[VALUE_COLUMN_X, TEXT_ROWS_START_Y - TEXT_ROW_HEIGHT * 3],
    color='black', alignText='left', anchorHoriz='left',
    height=TEXT_STIM_HEIGHT, wrapWidth=LABEL_WRAP_LENGTH)
modifiers_stim = visual.TextStim(win, units=unit_type, text=u'',
    pos=[VALUE_COLUMN_X, TEXT_ROWS_START_Y - TEXT_ROW_HEIGHT * 4],
    color='black', alignText='left', anchorHoriz='left',
    height=TEXT_STIM_HEIGHT, wrapWidth=LABEL_WRAP_LENGTH)
keypress_duration_stim = visual.TextStim(win, units=unit_type, text=u'',
    pos=[VALUE_COLUMN_X, TEXT_ROWS_START_Y - TEXT_ROW_HEIGHT * 5],
    color='black', alignText='left', anchorHoriz='left',
    height=TEXT_STIM_HEIGHT, wrapWidth=LABEL_WRAP_LENGTH)
all_pressed_stim = visual.TextStim(win, units=unit_type, text=u'',
    pos=[VALUE_COLUMN_X, TEXT_ROWS_START_Y - TEXT_ROW_HEIGHT * 6],
    color='black', alignText='left', anchorHoriz='left',
    height=TEXT_STIM_HEIGHT, wrapWidth=LABEL_WRAP_LENGTH)
event_type_stim = visual.TextStim(win, units=unit_type, text=u'',
    pos=[VALUE_COLUMN_X, TEXT_ROWS_START_Y - TEXT_ROW_HEIGHT * 7],
    color='black', alignText='left', anchorHoriz='left',
    height=TEXT_STIM_HEIGHT, wrapWidth=LABEL_WRAP_LENGTH)
psychopy_key_stim = visual.TextStim(win, units=unit_type, text=u'',
    pos=[VALUE_COLUMN_X, TEXT_ROWS_START_Y - TEXT_ROW_HEIGHT * 8],
    color='black', alignText='left', anchorHoriz='left',
    height=TEXT_STIM_HEIGHT,  wrapWidth=dw * 2)
ptb_key_stim = visual.TextStim(win, units=unit_type, text=u'',
    pos=[VALUE_COLUMN_X, TEXT_ROWS_START_Y - TEXT_ROW_HEIGHT * 9],
    color='black', alignText='left', anchorHoriz='left',
    height=TEXT_STIM_HEIGHT,  wrapWidth=dw * 2)

# Having all the stim to update / draw in a list makes drawing code
# more compact and reusable
STIM_LIST = [title_label, title2_label, key_text_label, char_label,
    modifiers_label, keypress_duration_label, all_pressed__label,
    event_type_label, psychopy_key_label, ptb_key_label,
    key_text_stim, char_stim, modifiers_stim, keypress_duration_stim,
    all_pressed_stim, event_type_stim, psychopy_key_stim, ptb_key_stim]

# Clear all events from the global and device level ioHub Event Buffers.

io.clearEvents('all')

QUIT_EXP = False
demo_timeout_start = core.getTime()

# Loop until we get a 'escape' key is pressed,
# or until 15 seconds passed since last keyboard event.

# Note that keyboard events can be compared to a string, matching when
# the event.key or .char == basestring value.
events = []
flip_time = demo_timeout_start = 0
while not 'q' in events and flip_time - demo_timeout_start < 15.0:
    for s in STIM_LIST:
        s.draw()
    flip_time = win.flip()

    events = keyboard.getKeys()

    for kbe in events:
        key_text_stim.text = kbe.key
        char_stim.text = kbe.char
        modifiers_stim.text = str(kbe.modifiers)

        psychopy_keys = event.getKeys()
        if psychopy_keys:
            psychopy_key_stim.text = psychopy_keys[0]
        elif kbe.type == "KEYBOARD_PRESS":
            psychopy_key_stim.text = ''

        ptb_keys = ptb_keyboard.getKeys(waitRelease=False)
        if ptb_keys:
            ptb_key_stim.text = ptb_keys[0].name

        all_pressed_stim.text = str(list(keyboard.state.keys()))

        if kbe.type == "KEYBOARD_PRESS":
            keypress_duration_stim.text = ''
        else:
            keypress_duration_stim.text = "%.6f" % kbe.duration

        event_type_stim.text = kbe.type
        demo_timeout_start = kbe.time

win.close()
core.quit()

# The contents of this file are in the public domain.



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Tests the round trip delay from when the experiment runtime requests
new events from the ioHub server to when a response with >=1 new
event is received and ready for use within the experiment script.

Only getEvent requests that return with at least one new event are used in
the calculated statistics to try and ensure the reported delay is measuring
the higher processing load case of new events being returned, vs. the
case of no new events being available.

At the end of the test, a MatPlotLib figure is displayed showing a
histogram of the round trip event request delays as well as two figures
representing the retrace onset detection stability of PsychoPy.
"""

from numpy import zeros
from scipy.stats import norm
from psychopy import visual
from psychopy.iohub import Computer, launchHubServer
from psychopy.iohub.constants import EventConstants
from collections import OrderedDict

totalEventRequestsForTest = 1000
numEventRequests = 0


def run():
    global numEventRequests
    # create fullscreen pyglet window at current resolution, as well as required resources / drawings
    psychoWindow, psychoStim = createPsychoGraphicsWindow()

    io = launchHubServer(window=psychoWindow, experiment_code='delay_test')
    io.devices.mouse.setPosition((0, 0))

    lastFlipTime = 0.0

    # create stats numpy arrays, set experiment process to high priority.
    # Init Results numpy array
    results = zeros((totalEventRequestsForTest, 3), dtype='f4')

    numEventRequests = 0

    # clear the ioHub event Buffer before starting the test.
    # This is VERY IMPORTANT, given an existing bug in ioHub.
    # You would want to do this before each trial started until the bug is fixed.
    io.clearEvents('all')

    # draw and flip to the updated graphics state.
    flipTime = drawAndFlipPsychoWindow(psychoStim, psychoWindow, io, None)
    ifi = flipTime - lastFlipTime
    lastFlipTime = flipTime

    # START TEST LOOP >>>>>>>>>>>>>>>>>>>>>>>>>>
    while numEventRequests < totalEventRequestsForTest:
        # send an Experiment Event to the ioHub server process
        io.sendMessageEvent("This is a test message %.3f" % flipTime)

        # check for any new events from any of the devices, and return the events list and the time it took to
        # request the events and receive the reply
        events, callDuration = checkForEvents(io)
        if events:
            # events were available
            results[numEventRequests][0] = callDuration  # ctime it took to get events from ioHub
            results[numEventRequests][1] = len(events)  # number of events returned
            results[numEventRequests][2] = ifi * 1000.0  # calculating inter flip interval.
            numEventRequests += 1  # incrementing tally counterfgh
            # draw and flip to the updated graphics state.

        flipTime = drawAndFlipPsychoWindow(psychoStim, psychoWindow, io, events)
        ifi = flipTime - lastFlipTime
        lastFlipTime = flipTime

    # END TEST LOOP <<<<<<<<<<<<<<<<<<<<<<<<<<

    psychoWindow.close()

    # plot collected delay and retrace detection results.
    plotResults(results)
    printResults(results)


def createPsychoGraphicsWindow():
    # create a window
    psychoStim = OrderedDict()
    psychoWindow = visual.Window((1920, 1080),
                                 monitor='default',
                                 units='pix',
                                 color=[128, 128, 128], colorSpace='rgb255',
                                 fullscr=True, allowGUI=False,
                                 screen=0
                                 )

    psychoWindow.setMouseVisible(False)

    fixation = visual.PatchStim(psychoWindow, size=25, pos=[0, 0], sf=0,
                                color=[-1, -1, -1], colorSpace='rgb')
    title = visual.TextStim(win=psychoWindow,
                            text="ioHub getEvents Delay Test", pos=[0, 125],
                            height=36, color=[1, .5, 0], colorSpace='rgb',
                            wrapWidth=800.0)

    instr = visual.TextStim(win=psychoWindow,
                            text='Move the mouse around, press keyboard keys and mouse buttons',
                            pos=[0, -125], height=32, color=[-1, -1, -1],
                            colorSpace='rgb', wrapWidth=800.0)

    psychoStim['static'] = visual.BufferImageStim(win=psychoWindow,
                                                  stim=(fixation, title, instr))
    psychoStim['grating'] = visual.PatchStim(psychoWindow,
                                             mask="circle", size=75, pos=[-100, 0],
                                             sf=.075)
    psychoStim['keytext'] = visual.TextStim(win=psychoWindow,
                                            text='key', pos=[0, 300], height=48,
                                            color=[-1, -1, -1], colorSpace='rgb',
                                            wrapWidth=800.0)
    psychoStim['mouseDot'] = visual.GratingStim(win=psychoWindow,
                                                tex=None, mask="gauss",
                                                pos=(0,0), size=(50, 50),
                                                color='purple')
    psychoStim['progress'] = visual.ShapeStim(win=psychoWindow,
                                              vertices=[(0, 0), (0, 0), (0, 0), (0, 0)],
                                              pos=(400, -300))

    return psychoWindow, psychoStim


def drawAndFlipPsychoWindow(psychoStim, psychoWindow, io, events):
    psychoStim['grating'].setPhase(0.05, '+')  # advance phase by 0.05 of a cycle
    currentPosition, currentDisplayIndex = io.devices.mouse.getPosition(return_display_index=True)

    if currentDisplayIndex == 0:
        currentPosition = (float(currentPosition[0]), float(currentPosition[1]))
        psychoStim['mouseDot'].setPos(currentPosition)

    if events:
        diff = totalEventRequestsForTest - numEventRequests
        v = psychoWindow.size[1] / 2.0 * diff / totalEventRequestsForTest
        vert = [[0, 0], [0, v], [2, v], [2, 0]]
        psychoStim['progress'].setVertices(vert)

        for r in events:
            if r.type is EventConstants.KEYBOARD_PRESS:  # keypress code
                psychoStim['keytext'].setText(r.key)

    [psychoStim[skey].draw() for skey in psychoStim]

    flipTime = psychoWindow.flip()
    return flipTime


def checkForEvents(io):
    # get the time we request events from the ioHub
    stime = Computer.getTime()
    r = io.getEvents()
    if r and len(r) > 0:
        # so there were events returned in the request, so include this getEvent request in the tally
        etime = Computer.getTime()
        dur = etime - stime
        return r, dur * 1000.0
    return None, None


def plotResults(results):
    #### calculate stats on collected data and draw some plots ####
    from matplotlib.pyplot import axis, title, xlabel, hist, grid, show, ylabel, plot
    import pylab

    durations = results[:, 0]
    flips = results[1:, 2]

    dmean = durations.mean()
    dstd = durations.std()

    fmean = flips.mean()
    fstd = flips.std()

    pylab.figure(figsize=(7, 5))
    pylab.subplot(1, 3, 1)

    # the histogram of the delay data
    n, bins, patches = hist(durations, 50, facecolor='blue', alpha=0.75)
    # add a 'best fit' line
    y = norm.pdf(bins, dmean, dstd)
    plot(bins, y, 'r--', linewidth=1)
    xlabel('ioHub getEvents Delay')
    ylabel('Percentage')
    title('ioHub Event Delays (msec):\n' + r'$\ \mu={0:.3f},\ \sigma={1:.3f}$'.format(dmean, dstd))
    axis([0, durations.max() + 1.0, 0, 25.0])
    grid(True)

    # graphs of the retrace data ( taken from retrace example in psychopy demos folder)
    intervalsMS = flips
    m = fmean
    sd = fstd
    distString = "Mean={0:.1f}ms,    s.d.={1:.1f},    99%CI={2:.1f}-{3:.1f}".format(
        m, sd, m - 3 * sd, m + 3 * sd)
    nTotal = len(intervalsMS)
    nDropped = sum(intervalsMS > (1.5 * m))
    droppedString = "Dropped/Frames = {0:d}/{1:d} = {2:0.2f}%".format(
        nDropped, nTotal, int(nDropped) / float(nTotal))

    pylab.subplot(1, 3, 2)

    # plot the frameintervals
    pylab.plot(intervalsMS, '-')
    pylab.ylabel('t (ms)')
    pylab.xlabel('frame N')
    pylab.title(droppedString)

    pylab.subplot(1, 3, 3)
    pylab.hist(intervalsMS, 50, histtype='stepfilled')
    pylab.xlabel('t (ms)')
    pylab.ylabel('n frames')
    pylab.title(distString)
    show()


def printResults(results):
    durations = results[:, 0]
    dmean = durations.mean()
    dstd = durations.std()
    print("ioHub getEvent Delays:")
    print("\tMEAN: ", dmean)
    print("\tSDEV: ", dstd)


if __name__ == "__main__":
    run()



"""Tests for psychopy.hardware.emulator"""

# covers most lines of code but does not test all possible logic

# py.test -k emulator --cov-report term-missing --cov hardware/emulator.py tests/test_hardware

import os, sys
import pytest

from psychopy import visual, core, event
from psychopy.hardware import emulator
from psychopy.hardware.emulator import *

# launchScan sound is not tested, nor included coverage

travis = bool("{}".format(os.environ.get('TRAVIS')).lower() == 'true')

BASE_MR_SETTINGS = {
    'TR': 0.5,    # duration (sec) per volume
    'volumes': 3, # number of whole-brain 3D volumes / frames
    'sync': '5',  # character to use as the sync timing event; assumed to come at start of a volume
    'skip': 1     # number of volumes lacking a sync pulse at start of scan (for T1 stabilization)
    }

@pytest.mark.emulator
class TestLaunchScan():
    '''A base class to test launchScan with different MR_settings'''
    def setup(self):
        self.win = visual.Window(fullscr=False, autoLog=False)
        self.globalClock = core.Clock()
        self.MR_settings = BASE_MR_SETTINGS.copy()

    def test_launch_scan(self):
        '''Test that launchScan successfully adds sync keys to the buffer.'''
        MR_settings = self.MR_settings

        # Initialize onsets with 0; the first onset is used to 0 the clock and is not reported by launchScan
        onsets = [0.0]
        sync_key = MR_settings['sync']

        # Call launchScan with MR Settings from test classes, in 'Test' mode with a short timeout.
        vol = launchScan(self.win, MR_settings, globalClock=self.globalClock,
                         mode='Test', wait_timeout=5, log=False)
        assert 0 < self.globalClock.getTime() < .05  # should get zeroed upon launch

        duration = MR_settings['volumes'] * MR_settings['TR']
        while self.globalClock.getTime() < duration:
            allKeys = event.getKeys(timeStamped=True)
            # detect sync or infer it should have happened:
            for key_tuple in allKeys:
                if key_tuple[0] == sync_key:
                    vol += 1
                    onsets.append(key_tuple[1])
        assert vol == MR_settings['volumes'] == len(onsets)

    def test_no_mode(self):
        #pytest.skip()
        event.clearEvents()
        ResponseEmulator(simResponses=[(0.5, 'escape')]).run()
        vol = launchScan(self.win, BASE_MR_SETTINGS.copy(), globalClock=self.globalClock,
                         wait_timeout=1, log=False)
        # no mode, so a RatingScale will be displayed; return to select value
        # the key event is not being detected here
        core.wait(1, 0)
        #event._onPygletKey(symbol='escape', modifiers=None, emulated=True)
        #core.wait(1, 0)

    def test_sync_generator(self):
        with pytest.raises(ValueError):
            s = SyncGenerator(TR=0.01)
        s = SyncGenerator(TR=0.1)
        s.start()
        s.stop()

    def test_response_emulator(self):
        # logs error but does not raise:
        ResponseEmulator(simResponses=[(.1, 0.123)]).run()

        r = ResponseEmulator()
        r.start()
        r.stop()
        core.wait(.1, 0)

    def test_misc(self):
        if travis:
            pytest.skip()
        MR_settings = BASE_MR_SETTINGS.copy()
        MR_settings.update({'sync': 'equal'})
        vol = launchScan(self.win, MR_settings, globalClock=self.globalClock,
                         simResponses=[(0.1,'a'), (.2, 1)],
                         mode='Test', log=False)
        # test replace missing defaults:
        min_MR_settings = {
            'TR': 0.2,    # duration (sec) per volume
            'volumes': 3}
        vol = launchScan(self.win, min_MR_settings, globalClock=self.globalClock,
                         mode='Test', log=False)
        core.wait(1, 0)


    def test_wait_timeout(self):
        '''Ensure that the wait_timeout happily rejects bad values.'''
        with pytest.raises(RuntimeError):
            vol = launchScan(self.win, BASE_MR_SETTINGS, wait_timeout=.1,
                             mode='Scan', log=False)
        with pytest.raises(ValueError):
            vol = launchScan(self.win, self.MR_settings, wait_timeout='cant_coerce_me!',
                             mode='Test', log=False)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Part of the PsychoPy library
# Copyright (C) 2002-2018 Jonathan Peirce (C) 2019-2022 Open Science Tools Ltd.
# Distributed under the terms of the GNU General Public License (GPL).
import sys

import wx
from .. import utils
from . import functions
import re
from pathlib import Path
from psychopy.tools.versionchooser import _translate
from ...tools.stringtools import valid_proj_name


class SyncDialog(wx.Dialog):
    def __init__(self, parent, project):
        wx.Dialog.__init__(self, parent, title=_translate("Syncing project..."))
        self.project = project
        # Setup sizer
        self.sizer = wx.BoxSizer(wx.VERTICAL)
        self.SetSizer(self.sizer)
        # Create status panel
        self.status = InfoStream(self, id=wx.ID_ANY, size=(-1, -1),
                                 value=_translate("Synchronising..."),
                                 style=wx.TE_READONLY | wx.TE_MULTILINE)
        self.sizer.Add(self.status, border=6, proportion=1, flag=wx.ALL | wx.EXPAND)
        # Setup button sizer
        self.btnSizer = wx.BoxSizer(wx.HORIZONTAL)
        self.sizer.Add(self.btnSizer, border=6, flag=wx.ALL | wx.EXPAND)
        self.btnSizer.AddStretchSpacer(1)
        # Add buttons
        self.OKbtn = wx.Button(self, label=_translate("OK"), id=wx.ID_OK)
        self.OKbtn.Disable()
        self.btnSizer.Add(self.OKbtn, border=3, flag=wx.LEFT | wx.ALIGN_CENTER_VERTICAL)
        # Layout
        self.Layout()
        self.Show()

    def sync(self):
        # If there's no user yet, login
        if self.project.session.user is None:
            functions.logInPavlovia(self)
        # Do sync
        self.project.sync(self.status)
        self.OKbtn.Enable()


class InfoStream(wx.TextCtrl):
    def __init__(self, parent, id, size,
                 value="Synchronising...",
                 style=wx.TE_READONLY | wx.TE_MULTILINE):
        wx.TextCtrl.__init__(self, parent, id,
                             size=size, value=value, style=style)

    def clear(self):
        self.SetValue("")

    def write(self, text):
        if type(text) == bytes:
            text = text.decode('utf-8')
        # Sanitize text (remove sensitive info like oauth keys)
        text = utils.sanitize(text)
        # Show
        self.SetValue(self.GetValue() + text)
        wx.Yield()


class CreateDlg(wx.Dialog):
    # List of folders which are invalid paths for a pavlovia project
    invalidFolders = [Path.home() / 'Desktop',
                      Path.home() / 'My Documents',
                      Path.home() / 'Documents']

    def __init__(self, parent, user, name="", path=""):
        wx.Dialog.__init__(self, parent=parent,
                           title=_translate("New project..."),
                           size=(500, 200), style=wx.DEFAULT_DIALOG_STYLE | wx.CLOSE_BOX)
        # If there's no user yet, login
        if user is None:
            user = functions.logInPavlovia(self)

        self.user = user
        self.session = parent.session
        self.project = None

        # Setup sizer
        self.frame = wx.BoxSizer(wx.VERTICAL)
        self.SetSizer(self.frame)
        self.sizer = wx.BoxSizer(wx.VERTICAL)
        self.frame.Add(self.sizer, border=6, proportion=1, flag=wx.ALL | wx.EXPAND)

        # Name label
        self.nameLbl = wx.StaticText(self, label=_translate("Project name:"))
        self.sizer.Add(self.nameLbl, border=3, flag=wx.ALL | wx.EXPAND)
        # Name ctrls
        self.nameSizer = wx.BoxSizer(wx.HORIZONTAL)
        self.sizer.Add(self.nameSizer, border=3, flag=wx.ALL | wx.EXPAND)
        # URL prefix
        self.nameRootLbl = wx.StaticText(self, label="pavlovia.org /")
        self.nameSizer.Add(self.nameRootLbl, border=3, flag=wx.ALL | wx.ALIGN_CENTER_VERTICAL)
        # Namespace ctrl
        self.namespaceCtrl = wx.Choice(self, choices=[user['username']] + user.session.listUserGroups(namesOnly=True))
        self.namespaceCtrl.SetStringSelection(user['username'])
        self.nameSizer.Add(self.namespaceCtrl, border=3, flag=wx.ALL | wx.EXPAND)
        # Slash
        self.slashLbl = wx.StaticText(self, label="/")
        self.nameSizer.Add(self.slashLbl, border=3, flag=wx.ALL | wx.ALIGN_CENTER_VERTICAL)
        # Name ctrl
        self.nameCtrl = wx.TextCtrl(self, value=str(name))
        self.nameCtrl.Bind(wx.EVT_TEXT, self.validate)
        self.nameSizer.Add(self.nameCtrl, border=3, proportion=1, flag=wx.ALL | wx.EXPAND)

        # Local root label
        self.rootLbl = wx.StaticText(self, label=_translate("Project folder:"))
        self.sizer.Add(self.rootLbl, border=3, flag=wx.ALL | wx.EXPAND)
        # Local root ctrl
        self.rootCtrl = utils.FileCtrl(self, value=str(path), dlgtype="dir")
        self.rootCtrl.Bind(wx.EVT_FILEPICKER_CHANGED, self.validate)
        self.sizer.Add(self.rootCtrl, border=3, flag=wx.ALL | wx.EXPAND)

        # Add dlg buttons
        self.sizer.AddStretchSpacer(1)
        self.btnSizer = wx.BoxSizer(wx.HORIZONTAL)
        self.sizer.Add(self.btnSizer, border=3, flag=wx.ALL | wx.EXPAND)
        self.btnSizer.AddStretchSpacer(1)
        # OK button
        self.OKbtn = wx.Button(self, id=wx.ID_OK, label=_translate("OK"))
        self.OKbtn.Bind(wx.EVT_BUTTON, self.submit)
        # CANCEL button
        self.CANCELbtn = wx.Button(self, id=wx.ID_CANCEL, label=_translate("Cancel"))
        # Add dlg buttons in OS appropriate order
        if sys.platform == "win32":
            btns = [self.OKbtn, self.CANCELbtn]
        else:
            btns = [self.CANCELbtn, self.OKbtn]
        self.btnSizer.Add(btns[0], border=3, flag=wx.ALL)
        self.btnSizer.Add(btns[1], border=3, flag=wx.ALL)

        self.Layout()
        self.validate()

    def validate(self, evt=None):
        # Test name
        name = self.nameCtrl.GetValue()
        nameValid = bool(valid_proj_name.fullmatch(name))
        # Test path
        path = Path(self.rootCtrl.GetValue())
        pathValid = path.is_dir() and path not in self.invalidFolders
        # Combine
        valid = nameValid and pathValid
        # Enable/disable Okay button
        self.OKbtn.Enable(valid)

        return valid

    def submit(self, evt=None):
        self.project = self.session.createProject(**self.GetValue())
        if self.project is not None:
            self.project.refresh()
            evt.Skip()

    def GetValue(self):
        return {
            "name": self.nameCtrl.GetValue(),
            "localRoot": self.rootCtrl.GetValue(),
            "namespace": self.namespaceCtrl.GetStringSelection()
        }